//3 BYTES PER PIXEL

/*----------------------------------------------------------------------------
:
:												BASIC GRAPHIC LIBRARY
:											Programmed by Manuel Guillen.
:
:			This Class can be used with any Video-Mode with True RGB colours
:           VESA Modes 0x112,
----------------------------------------------------------------------------*/
// MODEs 112h, 115h
// Res 640x480, 800x600    16.8M colours (8:8:8)
// Bytes per scan line: 2048,1920
// Win Granularity: 4, 16, 64 Kb
// Win Size: 64 Kb
// WinA Segment: A000
// WinB Segment: A000,0000
// Bits per pixel: 24
// RGB    8:8:8

#include "BGraph.hpp"
#include "Message.hpp"

#include <stdio.h>
#include <process.h>
#include <dos.h>
#include <conio.h>
#include <mem.h>
#include <alloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


// WARNING! when using function Write(char *, ...)
#define MAX_CHAR_IN_WRITE   300 // WARNING: attempt to write more
																// than MAX_CHAR_IN_WRITE caracters
																// will be disastrous

//#define BYTES_PER_SCAN_LINE  1920L
#define WIN_GRANULARITY      65536L
#define WIN_GRANULARITY_SHIFT 5


extern MessageHandler message;


// static global variables (already allocated for fastest execution)
// used by MSetPixel and SetPixel
unsigned int win_pos_a, new_win_pos_a;
unsigned int var_win_pos_b, var_new_win_pos_b;

unsigned int& win_pos_b=var_win_pos_b;
unsigned int& new_win_pos_b=var_new_win_pos_b;

unsigned long add;
unsigned char far *aux_pt;
unsigned int offset;

// GetWinAB: initialised with 0x0000 or 0x0001 depending on
//  availability of WinB for 'win_pos_b' when reading
unsigned int GetWinAB;

/*----------------------------------------------------------------------------
												CONTRUCTOR & DESTRUCTOR
----------------------------------------------------------------------------*/
//BGraph::BGraph(void) {}

//BGraph::~BGraph(void) {}

/*----------------------------------------------------------------------------
														InitBgraph
----------------------------------------------------------------------------*/
int BGraph::InitBGraph() {

	if(WindowABModel == USE_ONLY_WIN_A) {
		// make the variables for WinB point to the variables for WinA
      &win_pos_b = &win_pos_a;
      &new_win_pos_b = &new_win_pos_a;
		message.Message(TO_FILE, "GBraph::InitGraph  'WinB Not Supported: win_pos_b = win_pos_a'");
		GetWinAB = 0x0000;
	}

	else {
		message.Message(TO_FILE, "BGraph::InitGraph  'WinB Supported: win_pos_a != win_pos_b'");
		GetWinAB = 0x0001;
	}

//	if(BytesPerPixel!=3) {
		message.Message(TO_FILE, "BGraph::InitGraph	 BytesPerPixel=%u", // Not Supported by BGraph",
			BytesPerPixel);
//		return VESA_ERROR;
//	}

	return VESA_OK;

}

/*----------------------------------------------------------------------------
														Macro Set Pixel
// It's not worth to implement:
// It would be a long piece of code and difficult to maintain
----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
															Set Pixel NoClip
----------------------------------------------------------------------------*/
void BGraph::SetPixelNoClip(int x, int y, RGB &rgb) {

	add = (BytesPerPixel * x + ((long)BytesPerScanLine * y));

	// check changes of memory window
	// call function only if the memory window is different
	new_win_pos_a = add / WinGranularity;     // Window position in video memory (in window granularity units)
	if (win_pos_a != new_win_pos_a) {
		win_pos_a = new_win_pos_a;
		_DX = win_pos_a;
		_BX = 0x0000;     // Select Super VGA video memory window A
		(*WinFuncAddr)();
	}


	offset = add - win_pos_a * WinGranularity;

	// if all 3 bytes (RGB) lay inside the Memory Window
	if(offset < 65534) {
		*((char far*)(WinAPointer + (offset++) ) ) = rgb.blue;
		*((char far*)(WinAPointer + (offset++) ) ) = rgb.green;
		*((char far*)(WinAPointer +  offset    ) ) = rgb.red;
	}

	// if the 3 bytes (RGB) lay at the border of the Memory Window
	else if(offset==65534) {
		*((char far*)WinAPointer+65534) = rgb.blue;   // almost last address of the Memory Window
		*((char far*)WinAPointer+65535) = rgb.green;  // last address of the Memory Window
		VESA::SetMemoryWindowA(++win_pos_a);        // Change to next Memory Page
		*((char far*)WinAPointer)       = rgb.red;    // first address of the Memory Window
	}
	// if the 3 bytes (RGB) lay at the border of the Memory Window
	else if(offset==65535) {
		*((char far*)WinAPointer+65535) = rgb.blue;   // last address of the Memory Window
		VESA::SetMemoryWindowA(++win_pos_a);				 // Change to next Memory Page
		*((char far*)WinAPointer)       = rgb.green;	// first address of the Memory Window
		*((char far*)WinAPointer+1)     = rgb.red;    // second address of the Memory Window
	}

}

/*----------------------------------------------------------------------------
														Get Pixel NoClip
----------------------------------------------------------------------------*/
RGB BGraph::GetPixelNoClip(int x, int y) {

	RGB rgb;

	add = (BytesPerPixel * x + ((long)BytesPerScanLine * y));

	// check changes of memory window
	// call function only if the memory window is different
	new_win_pos_b = add / WinGranularity;     // Window position in video memory (in window granularity units)
	if (win_pos_b != new_win_pos_b) {
		win_pos_b = new_win_pos_b;
		_DX = win_pos_b;
		_BX = GetWinAB;  // Select Super VGA video memory window A or B depending on availability of WinB
		(*WinFuncAddr)();
	}

	offset = add - (unsigned long)win_pos_b * WinGranularity;

	// if all 3 bytes (RGB) lay inside the Memory Window
	if(offset < 65534) {
		rgb.blue  = *((char far*)(WinBPointer + (offset++) ) );
		rgb.green = *((char far*)(WinBPointer + (offset++) ) );
		rgb.red   = *((char far*)(WinBPointer +  offset    ) );
	}

	// if the 3 bytes (RGB) lay at the border of the Memory Window
	else if(offset==65534) {
		rgb.blue  = *((char far*)WinBPointer+65534);   // almost last address of the Memory Window
		rgb.green = *((char far*)WinBPointer+65535);  // last address of the Memory Window
		SetMemoryWindowB(++win_pos_b);        // Change to next Memory Page
		rgb.red   = *((char far*)WinBPointer);          // first address of the Memory Window
	}
	// if the 3 bytes (RGB) lay at the border of the Memory Window
	else if(offset==65535) {
		rgb.blue  = *((char far*)WinBPointer+65535); // last address of the Memory Window
		SetMemoryWindowB(++win_pos_b);						// Change to next Memory Page
		rgb.green = *((char far*)WinBPointer);			// first address of the Memory Window
		rgb.red   = *((char far*)WinBPointer+1);      // second address of the Memory Window
	}

	return rgb;

}

/*----------------------------------------------------------------------------
																SetHLineNoClip
----------------------------------------------------------------------------*/
void BGraph::SetHLineNoClip(unsigned char *source, int x, int y, int dx) {
	unsigned int i_source=0;

	add = (BytesPerPixel * x + ((long)BytesPerScanLine * y));
//	add = (x + x + x + ((unsigned long)BytesPerScanLine * y));

	new_win_pos_a = add / WinGranularity;     // Window position in video memory (in window granularity units)
	if (win_pos_a != new_win_pos_a) {
		win_pos_a = new_win_pos_a;
		_DX = win_pos_a;
		_BX = 0x0000;     // Select Super VGA video memory window A
		(*WinFuncAddr)();
	}
	offset = add - win_pos_a * WinGranularity;

	// if all dx*BytesPerPixel bytes (RGB) lay inside the Memory Window
	if((long)offset+BytesPerPixel*dx < 65534) {
		for(; dx>0; dx--) {
			*((char far*)(WinAPointer + (offset++) ) ) = source[i_source++];
			*((char far*)(WinAPointer + (offset++) ) ) = source[i_source++];
			*((char far*)(WinAPointer + (offset++) ) ) = source[i_source++];
			if(BytesPerPixel==4) offset++;    // 4BBP
		}
	}

	// if the BytesPerPixel*dx bytes (RGB) cross the border of the Memory Window
	else	{
		unsigned long loffset = offset;
		while(dx>0) {

			while(dx>0 && loffset<65534L && loffset!=65536L) {
				*((char far*)(WinAPointer + (loffset++) ) ) = source[i_source++];
				*((char far*)(WinAPointer + (loffset++) ) ) = source[i_source++];
				*((char far*)(WinAPointer + (loffset++) ) ) = source[i_source++];
				dx--;
				if(BytesPerPixel==4) loffset++;       // 4BBP
			}

			if(loffset == 65536L) {
				VESA::SetMemoryWindowA(++win_pos_a);				 // Change to next Memory Page
				loffset = 0;
			}

			else if(loffset==65534L  && dx>0) {
				*((char far*)WinAPointer+65534) = source[i_source++];   // almost last address of the Memory Window
				*((char far*)WinAPointer+65535) = source[i_source++];  // last address of the Memory Window
				VESA::SetMemoryWindowA(++win_pos_a);        // Change to next Memory Page
				*((char far*)WinAPointer)       = source[i_source++];    // first address of the Memory Window
				loffset=1; if(BytesPerPixel==4) loffset++;
				dx--;
			}

			else if(loffset==65535L && dx>0) {
				*((char far*)WinAPointer+65535) = source[i_source++];   // last address of the Memory Window
				VESA::SetMemoryWindowA(++win_pos_a);				 // Change to next Memory Page
				*((char far*)WinAPointer)       = source[i_source++];	// first address of the Memory Window
				*((char far*)WinAPointer+1)     = source[i_source++];    // second address of the Memory Window
				loffset=2; if(BytesPerPixel==4) loffset++;
				dx--;
			}

		} // while dx>0

	} // else

}


/*----------------------------------------------------------------------------
															GetHLineNoClip
----------------------------------------------------------------------------*/
void BGraph::GetHLineNoClip(unsigned char *dest, int x, int y, int dx) {
	unsigned int i_dest=0;

	add = (BytesPerPixel * x + ((long)BytesPerScanLine * y));

	new_win_pos_b = add / WinGranularity;     // Window position in video memory (in window granularity units)
	if (win_pos_b != new_win_pos_b) {
		win_pos_b = new_win_pos_b;
		_DX = win_pos_b;
		_BX = GetWinAB;  // Select Super VGA video memory window A or B depending on availability of WinB
		(*WinFuncAddr)();
	}

	offset = add - win_pos_b * WinGranularity;

	// if all dx*BytesPerPixel bytes (RGB) lay inside the Memory Window
	if((long)offset+BytesPerPixel*dx < 65534) {
		for(; dx>0; dx--) {
			dest[i_dest++] = *((char far*)(WinBPointer + (offset++) ) );
			dest[i_dest++] = *((char far*)(WinBPointer + (offset++) ) );
			dest[i_dest++] = *((char far*)(WinBPointer + (offset++) ) );
			if(BytesPerPixel==4) offset++;    // 4BBP
		}
	}

	// if the BytesPerPixel*dx bytes (RGB) cross the border of the Memory Window
	else	{
		unsigned long loffset = offset;
		while(dx>0) {

			while(dx>0 && loffset<65534L && loffset!=65536L) {
				dest[i_dest++] = *((char far*)(WinBPointer + (loffset++) ) );
				dest[i_dest++] = *((char far*)(WinBPointer + (loffset++) ) );
				dest[i_dest++] = *((char far*)(WinBPointer + (loffset++) ) );
				if(BytesPerPixel==4) loffset++;       // 4BBP
				dx--;
			}

			if(loffset == 65536L) {
				SetMemoryWindowB(++win_pos_b);				 // Change to next Memory Page
				loffset = 0;
			}

			else if(loffset==65534L && dx>0) {
				dest[i_dest++] = *((char far*)WinBPointer+65534);   // almost last address of the Memory Window
				dest[i_dest++] = *((char far*)WinBPointer+65535);  // last address of the Memory Window
				SetMemoryWindowB(++win_pos_b);        // Change to next Memory Page
				dest[i_dest++] = *((char far*)WinBPointer);  // first address of the Memory Window
				loffset=1; if(BytesPerPixel==4) loffset++;       // 4BBP
				dx--;
			}

			else if(loffset==65535L && dx>0) {
				dest[i_dest++] = *((char far*)WinBPointer+65535);   // last address of the Memory Window
				SetMemoryWindowB(++win_pos_b);				 // Change to next Memory Page
				dest[i_dest++] = *((char far*)WinBPointer);	// first address of the Memory Window
				dest[i_dest++] = *((char far*)WinBPointer+1);    // second address of the Memory Window
				loffset=2;if(BytesPerPixel==4) loffset++;       // 4BBP
				dx--;
			}

		} // while dx>0

	} // else

}



void BGraph::Cls(RGB rgb) {
	int i;
	int wp=0;

	int increment = 65536L/WinGranularity;
	int end_i = (unsigned long)BytesPerScanLine * (unsigned long)YMax / 65536L;
	int last_win_pos_a = win_pos_a;

	for(i=0; i<=end_i; i++) {
		_DX = wp;
		_BX = 0x0000;     // Select Super VGA video memory window A
		(*WinFuncAddr)();
		wp += increment;

		memset((void far *)WinAPointer, 0, 32768);
		memset((void far *)(WinAPointer+32768), 0, 32768);
	}

	GotoXY(0,0);

	win_pos_a = last_win_pos_a;
	SetMemoryWindowA(win_pos_a);

}