#include "GraphDev.hpp"
#include "Message.hpp"
#include "Pool.hpp"
#include "Screen.hpp"

#include "CLIP.hpp"

#include <stdio.h>
#include <process.h>
#include <dos.h>
#include <conio.h>
#include <mem.h>
#include <alloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


extern MessageHandler message;
extern Pool pool;
extern Screen screen;

// static members
char far *GraphicDevice::character_set_ptr=NULL;
unsigned int GraphicDevice::ScanBufferBytesPerPixel=0;


/*----------------------------------------------------------------------------
							CONSTRUCTOR
----------------------------------------------------------------------------*/
GraphicDevice::GraphicDevice(void) {

	text.colour_type = GRADIENT_SO_NE;

	text.colour =RGB(255,255,255);
	text.paper_colour = RGB(115,115,115);

	text.colour_from = RGB(255,128,0);
	text.paper_colour_from = RGB(115,115,115);

	text.colour_to = RGB(0,125,255);
	text.paper_colour_to = RGB(115,115,115);

	print_on_screen = 0;

	GraphicDevice::GotoXY(0,0);

}

GraphicDevice::~GraphicDevice() {
}

void GraphicDevice::ExtendScanBuffer(unsigned int x_size) {

	pool.Extend((unsigned int)ScanBufferBytesPerPixel * x_size);

}



/*----------------------------------------------------------------------------
CLIPPING TO BE TESTED
----------------------------------------------------------------------------*/

void GraphicDevice::SetPixel(int x, int y,	RGB &rgb) {
	if(x>=0 && x<XMax && y>=0 && y<=YMax) SetPixelNoClip(x, y, rgb);
}

/*----------------------------------------------------------------------------
CLIPPING TO BE TESTED
----------------------------------------------------------------------------*/
RGB GraphicDevice::GetPixel(int x, int y) {
	if(x>=0 && x<XMax && y>=0 && y<=YMax) return GetPixelNoClip(x, y);
	else return RGB(0,0,0);
}

/*----------------------------------------------------------------------------
CLIPPING TO BE TESTED
----------------------------------------------------------------------------*/
void GraphicDevice::SetHLine(unsigned char *source, int x, int y, int dx) {

	int x1 = x;

	CLIP_HLINE(x,y, dx, XMax,YMax);

	SetHLineNoClip(source + (x-x1)*3, x,y, dx);

}

/*----------------------------------------------------------------------------
CLIPPING TO BE TESTED
----------------------------------------------------------------------------*/
void GraphicDevice::GetHLine(unsigned char *dest, int x, int y, int dx) {

	int x1=x;

	CLIP_HLINE(x,y,dx,XMax,YMax)

	GetHLineNoClip(dest + (x-x1)*3, x,y, dx);
}


/*----------------------------------------------------------------------------
																HLine
CLIPPING TESTED
----------------------------------------------------------------------------*/
void GraphicDevice::HLine(int x0, int y0, int dx, RGB& rgb) {

	CLIP_HLINE(x0,y0, dx, XMax,YMax);

	unsigned int i=0;
	int dx2=dx;

	for(;dx>=0; dx--) {
		pool.buffer[i++]=rgb.blue;
		pool.buffer[i++]=rgb.green;
		pool.buffer[i++]=rgb.red;
	}

	SetHLine(pool.buffer, x0,y0, dx2);

}

/*----------------------------------------------------------------------------
																VLine
CLIPPING TESTED
----------------------------------------------------------------------------*/
void GraphicDevice::VLine(int x, int y, int dy, RGB& rgb) {

	CLIP_HLINE(y,x, dy, YMax,XMax);

	if(dy>0) while(dy--) SetPixelNoClip(x,y++, rgb);
	else while(dy++) SetPixelNoClip(x,y--, rgb);

}

/*----------------------------------------------------------------------------
																Line
----------------------------------------------------------------------------*/
void GraphicDevice::Line(int x1, int y1, int x2, int y2, RGB& rgb) {

	int i, deltax, deltay, numpixels,
		d, dinc1, dinc2,
		x, xinc1, xinc2,
		y, yinc1, yinc2;

	if(y1>y2) {
		i=x1; x1=x2; x2=i;
		i=y1; y1=y2; y2=i;
	}

	// Calculate deltax and deltay for initialisation
//	deltax = abs(x2 - x1);
	deltax = x2 - x1; if(deltax<0) deltax = -deltax;

//	deltay = abs(y2 - y1);
	deltay = y2 - y1; if(deltay<0) deltay = -deltay;

	// Initialize all vars based on which is the independent variable }
	if(deltax >= deltay) {
		// x is independent variable
		numpixels = deltax + 1;
		d = (2 * deltay) - deltax;
		dinc1 = deltay << 1;
		dinc2 = (deltay - deltax) << 1;
		xinc1 = 1;
		xinc2 = 1;
		yinc1 = 0;
		yinc2 = 1;
	}
	else {
			// y is independent variable
		numpixels = deltay + 1;
		d = (2 * deltax) - deltay;
		dinc1 = deltax << 1;
		dinc2 = (deltax - deltay) << 1;
		xinc1 = 0;
		xinc2 = 1;
		yinc1 = 1;
		yinc2 = 1;
	}

	// Make sure x and y move in the right directions
	if(x1 > x2) {
		xinc1 = - xinc1;
		xinc2 = - xinc2;
	}

	if(y1 > y2) {
		yinc1 = - yinc1;
		yinc2 = - yinc2;
	}

	// Start drawing at <x1, y1>
	x = x1;
	y = y1;

	// Draw the pixels
	for(i=1; i < numpixels; i++) {
		if(x>=0 && x<XMax && y>=0 && y<=YMax) SetPixelNoClip(x, y, rgb);
		if(d<0) {
			d = d + dinc1;
			x = x + xinc1;
			y = y + yinc1;
		}
		else {
			d = d + dinc2;
			x = x + xinc2;
			y = y + yinc2;
		}
	}


}


/*----------------------------------------------------------------------------
														GetBlock
   max size 65530 bytes!!
	memory format:
		header:
			int dx,
			int dy,
		data:
			char r,g,b,
			char r,g,b,...

// BlockSize = (dx * dy + 4)  bytes    (use macro BLOCK_SIZE(DX,DY) )
----------------------------------------------------------------------------*/
void GraphicDevice::GetBlock(unsigned char *buffer, int x, int y, int dx, int dy) {

	unsigned long add=4;

	buffer[0] = dx%256;
	buffer[1] = dx/256;
	buffer[2] = dy%256;
	buffer[3] = dy/256;

	for(; dy; dy--) {
		GetHLine(buffer+add, x, y++, dx);
		add += dx * 3;
	}

}

/*----------------------------------------------------------------------------
												SetBlock
	max size 65530 bytes!!

// CLIPPING TO BE TESTED
----------------------------------------------------------------------------*/
void GraphicDevice::SetBlock(unsigned char *buffer, int x, int y) {

	int dx,dy;

	dx = buffer[0];
	dx+= buffer[1]<<8;
	dy = buffer[2];
	dy+= buffer[3]<<8;

	//CLIP_BLOCK(x,y, dx,dy, XMax, YMax);
/*	int x_org=0;
	int y_org=0;
	CLIP_BLOCK(x_org, y_org, x, y, dx,dy, XMax, YMax);
*/
	unsigned long add = 4;// + x_org*3 + y_org*dx*3;

	for(; dy; dy--) {
		SetHLine/*NoClip*/(buffer+add, x, y++, dx);
		add += dx * 3;
	}
}


/*----------------------------------------------------------------------------
														Copy Block
----------------------------------------------------------------------------*/
void GraphicDevice::CopyBlock(
	int x_dest, int y_dest,
	GraphicDevice &gd_org,
	int x_org, int y_org,  // default: (x,y)=(0,0)
	int dx, int dy) {      // default: (dx,dy)=(-1,-1) then use max size

	// default meanings
	if(dx==-1) {
		x_org=0;
		y_org=0;
		dx = gd_org.XMax;
		dy = gd_org.YMax;
	}


// clipping
	int x_aux = 0;
	int y_aux = 0;

	// clip org source
	CLIP_BLOCK(x_aux, y_aux, x_org, y_org, dx, dy,  gd_org.XMax, gd_org.YMax);
	//GotoXY(0,180);

	//message.Message(TO_SCREEN,"x=%i y=%i dx=%i dy=%i", x_org, y_org, dx,dy);
	// reposition x_dest,y_dest
	x_dest += x_aux;
	y_dest += y_aux;

	CLIP_BLOCK(x_org, y_org, x_dest, y_dest, dx,dy, XMax, YMax);

	int y;
	if(y_dest > y_org) {
		for(y=dy-1; y>=0; y--) {
			gd_org.GetHLineNoClip(pool.buffer, x_org, y_org+y, dx);
			SetHLineNoClip(pool.buffer, x_dest, y_dest+y, dx);
		}
	}
	else {
		for(y=0; y < dy; y++) {
			gd_org.GetHLineNoClip(pool.buffer, x_org, y_org+y, dx);
			SetHLineNoClip(pool.buffer, x_dest, y_dest+y, dx);
		}
	}

}

/*----------------------------------------------------------------------------
														Copy Block Zoom
----------------------------------------------------------------------------*/
void GraphicDevice::CopyBlockZoom(
	int x_dest, int y_dest,
	GraphicDevice &gd_org,
	int x_org, int y_org,
	int dx, int dy,
	int zoom)
{
	// no zoom:
	if(zoom >= -1 && zoom <= 1) {
		GraphicDevice::CopyBlock(x_dest, y_dest, gd_org, x_org, y_org, dx, dy);
		return;
	}

	// zoom out:
	if(zoom < -1) {
		copy_block_zoom_neg(x_dest,y_dest, gd_org, x_org,y_org, dx,dy, zoom);
		return;
	}

	// zoom in:
	int i,y;

	// clipping
	int dx_ini = dx;
	int dy_ini = dy;
	int x_dest_ini = x_dest;
	int y_dest_ini = y_dest;

	int x_aux = 0;
	int y_aux = 0;

	int dx_aux = (dx+zoom-1)/zoom;
	int dy_aux = (dy+zoom-1)/zoom;

	// clip org source
	if(x_org<0 || y_org<0 || x_org+dx_aux > gd_org.XMax || y_org+dy_aux > gd_org.YMax) { // !I
		if(x_org+dx_aux <= 0) return;                                       // A
		if(y_org+dy_aux <= 0) return;                                       // B
		if(x_org >= gd_org.XMax) return;                                       // C
		if(y_org >= gd_org.YMax) return;                                       // D

		if(y_org<0) { dy_aux = dy_aux+y_org; y_aux=-y_org; y_org=0;}         // E
		if(y_org+dy_aux >= gd_org.YMax) { dy_aux = gd_org.YMax-y_org; }                  // F

		if(x_org<0) { dx_aux = dx_aux+x_org; x_aux=-x_org; x_org=0; }        // G
		if(x_org+dx_aux >= gd_org.XMax) { dx_aux = gd_org.XMax-x_org; }                  // H
	}

	dx = dx_aux*zoom;
	dy = dy_aux*zoom;

	// reposition x_dest,y_dest
	x_dest += x_aux*zoom; if(x_dest+dx > x_dest_ini+dx_ini) dx=dx_ini+(x_dest_ini-x_dest);
	y_dest += y_aux*zoom; if(y_dest+dy > y_dest_ini+dy_ini) dy=dy_ini+(y_dest_ini-y_dest);

	CLIP_BLOCK(x_org, y_org, x_dest, y_dest, dx,dy, XMax, YMax);

//
	RGB *rgb = (RGB *)pool.buffer;
	int dx_zoom = (dx+zoom-1)/zoom;

	for(y=0; y<dy; ) {
		gd_org.GetHLineNoClip(pool.buffer, x_org, y_org+y/zoom, dx_zoom);

		for(i=dx-1; i>=0; i--) rgb[i]=rgb[i/zoom];

		for(i=0; i<zoom && y<dy; i++,y++) {
			SetHLine(pool.buffer, x_dest, y_dest+y, dx);
		}
	}

}



/*----------------------------------------------------------------------------
														Copy Block Zoom < 0
----------------------------------------------------------------------------*/
void GraphicDevice::copy_block_zoom_neg(
	int x_dest, int y_dest,
	GraphicDevice &gd_org,
	int x_org, int y_org,
	int dx, int dy,
	int zoom)
{

	// zoom < -1

	zoom = -zoom;

	int i,y;

// clipping
	int dx_ini = dx;
	int dy_ini = dy;
	int x_dest_ini = x_dest;
	int y_dest_ini = y_dest;

	int x_aux = 0;
	int y_aux = 0;

	int dx_aux = dx*zoom;
	int dy_aux = dy*zoom;

	// clip org source
	if(x_org<0 || y_org<0 || x_org+dx_aux > gd_org.XMax || y_org+dy_aux > gd_org.YMax) { // !I
		if(x_org+dx_aux <= 0) return;                                       // A
		if(y_org+dy_aux <= 0) return;                                       // B
		if(x_org >= gd_org.XMax) return;                                       // C
		if(y_org >= gd_org.YMax) return;                                       // D

		if(y_org<0) { dy_aux = dy_aux+y_org; y_aux=-y_org; y_org=0;}         // E
		if(y_org+dy_aux >= gd_org.YMax) { dy_aux = gd_org.YMax-y_org; }                  // F

		if(x_org<0) { dx_aux = dx_aux+x_org; x_aux=-x_org; x_org=0; }        // G
		if(x_org+dx_aux >= gd_org.XMax) { dx_aux = gd_org.XMax-x_org; }                  // H
	}

	dx = (dx_aux+zoom-1)/zoom;
	dy = (dy_aux+zoom-1)/zoom;

	// reposition x_dest,y_dest
	x_dest += x_aux/zoom; if(x_dest+dx > x_dest_ini+dx_ini) dx=dx_ini+(x_dest_ini-x_dest);
	y_dest += y_aux/zoom; if(y_dest+dy > y_dest_ini+dy_ini) dy=dy_ini+(y_dest_ini-y_dest);


	CLIP_BLOCK(x_org, y_org, x_dest, y_dest, dx,dy, XMax, YMax);

//
	RGB *rgb = (RGB *)pool.buffer;
	int dx_zoom = dx*zoom;

	for(y=0; y<dy; y++) {
		gd_org.GetHLineNoClip(pool.buffer, x_org, y_org+y*zoom, dx_zoom);

		for(i=0; i<dx; i++) rgb[i]=rgb[i*zoom];

		SetHLine(pool.buffer, x_dest, y_dest+y, dx);


	}

}



/*----------------------------------------------------------------------------
															Cls
----------------------------------------------------------------------------*/
void GraphicDevice::Cls(RGB rgb) {

	Bar(0,0, XMax,YMax, rgb);
	GotoXY(0,0);

}

/*----------------------------------------------------------------------------
														 Bar
----------------------------------------------------------------------------*/
void GraphicDevice::Bar(int x, int y, int dx, int dy, RGB rgb) {

	// fill pool.buffer
	unsigned int i;
	for(i=0; i<XMax*3; ) {
		pool.buffer[i++]=rgb.blue;
		pool.buffer[i++]=rgb.green;
		pool.buffer[i++]=rgb.red;
	}

/* for mouse considerations only SetHLine call (no call to Clip version)
	// if the Bar is insided the printable area
	if(x>=0 && x+dx <= XMax && y>=0 && y+dy <= YMax) {
		for(;dy;dy--) SetHLineNoClip(pool.buffer, x,y++,dx);
	}
	else {
*/
		for(;dy;dy--) SetHLine(pool.buffer, x,y++,dx);
/*		message.Message(TO_FILE,"GraphicDevice::Bar  Warning:'Bar out of range");
	}*/

}

/*----------------------------------------------------------------------------
														 Rectangle
----------------------------------------------------------------------------*/
void GraphicDevice::Rectangle(int x, int y, int dx, int dy, RGB rgb) {

	HLine (x,      y,      dx,  rgb);
	VLine        (x,      y,      dy,  rgb);
	HLine (x,      y+dy-1, dx,  rgb);
	VLine        (x+dx-1, y,      dy,  rgb);

}

void GraphicDevice::Rectangle(RGB rgb) {

	Rectangle(0,0, XMax,YMax, rgb);

}

/*----------------------------------------------------------------------------
								 Functions to output text on the screen
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void GraphicDevice::GotoXY(int x, int y) {
	where_x = x;
	where_y = y;
}


/*----------------------------------------------------------------------------
													 PrintF
	Same as printf (print with format)

	WARNING: attempt to write more than MAX_CHAR_IN_WRITE caracters
	will be disastrous

----------------------------------------------------------------------------*/
void GraphicDevice::PrintF(char *format, va_list argptr) {

//	va_list argptr;
	char string[MAX_CHAR_IN_WRITE];   // max number of characters to write
	int i, prev_i, len;
	int string_eq_n;

//	va_start(argptr, format);
	vsprintf(string, format, argptr);
//	va_end(argptr);

	len = strlen(string);

	prev_i = 0;
	for(i=0; i<len; i++) {
		string_eq_n = string[i]=='\n';
		if((string[i]=='\r') || string_eq_n) {
			string[i]=0;
			PrintString(string + prev_i);
			if(string_eq_n) {
				where_x = 0;
				where_y += CHAR_HEIGHT;
			}
			else where_x += CHAR_WIDTH * strlen(string+prev_i);
			prev_i = i+1;
		}
	}
	if(prev_i != i) {
		PrintString(string + prev_i);
		where_x += CHAR_WIDTH * strlen(string + prev_i);
	}


}


//----------------------------------------------------------------------------
//    PrintString
//----------------------------------------------------------------------------
void GraphicDevice::PrintString(char *string) {
	unsigned int dim, pixels_per_scan, bytes_per_scan;
	unsigned char *image;
	char far *ptr;
	unsigned char bit_mask;
	int c, b, len;
	long x,y;
	unsigned int index;

// get memory
	len = strlen(string);
	pixels_per_scan = len * CHAR_WIDTH;
	bytes_per_scan = pixels_per_scan * 3;

	dim = bytes_per_scan * CHAR_HEIGHT + 4;

	if((image = (unsigned char *) malloc(dim))==NULL) {
////////////////		CloseGraph();
////////////////		printf("Not enough memory to allocate buffer in function PrintString()\n");
		exit(1);  // terminate program if out of memory
	}

// create block

	image[0] = pixels_per_scan%256;
	image[1] = pixels_per_scan/256;
	image[2] = CHAR_HEIGHT;
	image[3] = 0;

	index=4;
	for(y=0; y < CHAR_HEIGHT; y++) {
//		index = 4 + y * bytes_per_scan;
		for(c=0; c<len; c++) {
			ptr = character_set_ptr + string[c] * 16 + y;

			bit_mask = 0x80;
			for(b=0; b < 8; b++) {
				image[index] = *ptr & bit_mask;
				image[index+1]=0;
				image[index+2]=0;
				index+=3;
				bit_mask = bit_mask >> 1;
			}
		}
	}

// colour block

	index=4;

	switch(text.colour_type) {

		case SOLID:
			for(y=0; y < CHAR_HEIGHT; y++) {
				for(x=0; x<pixels_per_scan; x++) {
					if(image[index]) {
						image[index] = text.colour.blue;
						image[index+1] = text.colour.green;
						image[index+2] = text.colour.red;
					}
					else {
						image[index] = text.paper_colour.blue;
						image[index+1] = text.paper_colour.green;
						image[index+2] = text.paper_colour.red;
					}
					index+=3;
				}
			}
			break;

		case GRADIENT_NS:
			for(y=0; y < CHAR_HEIGHT; y++) {
				for(x=0; x<pixels_per_scan; x++) {
		//			index = 4 + y * bytes_per_scan + x*3;
					if(image[index]) {
		//	image[index] = text.colour_from_b * (1-y/CHAR_HEIGHT) + text.colour_to_b * (y/CHAR_HEIGHT);
						image[index] = text.colour_from.blue -(text.colour_from.blue*y)/CHAR_HEIGHT +
							(text.colour_to.blue * y)/CHAR_HEIGHT;
						image[index+1] = text.colour_from.green -(text.colour_from.green*y)/CHAR_HEIGHT +
							(text.colour_to.green * y)/CHAR_HEIGHT;
						image[index+2] = text.colour_from.red -(text.colour_from.red*y)/CHAR_HEIGHT +
							(text.colour_to.red * y)/CHAR_HEIGHT;
					}
					index+=3;
				}
			}
			break;


		case GRADIENT_OE:
			for(y=0; y < CHAR_HEIGHT; y++) {
				for(x=0; x<pixels_per_scan; x++) {
		//			index = 4 + y * bytes_per_scan + x*3;
					if(image[index]) {
		//	image[index] = text.colour_from_b * (1-y/CHAR_HEIGHT) + text.colour_to_b * (y/CHAR_HEIGHT);
						image[index] = text.colour_from.blue -(text.colour_from.blue*x)/pixels_per_scan +
							(text.colour_to.blue * x)/pixels_per_scan;
						image[index+1] = text.colour_from.green -(text.colour_from.green*x)/pixels_per_scan +
							(text.colour_to.green * x)/pixels_per_scan;
						image[index+2] = text.colour_from.red -(text.colour_from.red*x)/pixels_per_scan +
							(text.colour_to.red * x)/pixels_per_scan;
					}
					index+=3;
				}
			}
			break;

		case GRADIENT_NO_SE:
			for(y=0; y < CHAR_HEIGHT; y++) {
				for(x=0; x<pixels_per_scan; x++) {
		//			index = 4 + y * bytes_per_scan + x*3;
					if(image[index]) {
		//	image[index] = text.colour_from_b * (1-y/CHAR_HEIGHT) + text.colour_to_b * (y/CHAR_HEIGHT);
						image[index] = ((text.colour_from.blue -(text.colour_from.blue*x)/pixels_per_scan +
							(text.colour_to.blue * x)/pixels_per_scan) + (text.colour_from.blue -(text.colour_from.blue*y)/CHAR_HEIGHT +
							(text.colour_to.blue * y)/CHAR_HEIGHT)) / 2;

						image[index+1] = ((text.colour_from.green -(text.colour_from.green*x)/pixels_per_scan +
							(text.colour_to.green * x)/pixels_per_scan) + (text.colour_from.green -(text.colour_from.green*y)/CHAR_HEIGHT +
							(text.colour_to.green * y)/CHAR_HEIGHT)) / 2;

						image[index+2] = ((text.colour_from.red -(text.colour_from.red*x)/pixels_per_scan +
							(text.colour_to.red * x)/pixels_per_scan) + (text.colour_from.red -(text.colour_from.red*y)/CHAR_HEIGHT +
							(text.colour_to.red * y)/CHAR_HEIGHT)) / 2;
					}
					index+=3;
				}
			}
			break;

		case GRADIENT_SO_NE:
			for(y=0; y < CHAR_HEIGHT; y++) {
				for(x=0; x<pixels_per_scan; x++) {
					if(image[index]) {
						image[index] = ((text.colour_from.blue -(text.colour_from.blue*x)/pixels_per_scan +
							(text.colour_to.blue * x)/pixels_per_scan) + (text.colour_from.blue -(text.colour_from.blue*(CHAR_HEIGHT-y))/CHAR_HEIGHT +
							(text.colour_to.blue * (CHAR_HEIGHT-y))/CHAR_HEIGHT)) / 2;

						image[index+1] = ((text.colour_from.green -(text.colour_from.green*x)/pixels_per_scan +
							(text.colour_to.green * x)/pixels_per_scan) + (text.colour_from.green -(text.colour_from.green*(CHAR_HEIGHT-y))/CHAR_HEIGHT +
							(text.colour_to.green * (CHAR_HEIGHT-y))/CHAR_HEIGHT)) / 2;

						image[index+2] = ((text.colour_from.red -(text.colour_from.red*x)/pixels_per_scan +
							(text.colour_to.red * x)/pixels_per_scan) + (text.colour_from.red -(text.colour_from.red*(CHAR_HEIGHT-y))/CHAR_HEIGHT +
							(text.colour_to.red * (CHAR_HEIGHT-y))/CHAR_HEIGHT)) / 2;
					}
					index+=3;
				}
			}
			break;

		default:
		message.Message(EXIT | TO_FILE | TO_SCREEN, "GraphicDevice::PrintString  'text.colour_type not valid'");

	}

// print block

	if(print_on_screen)
		screen.SetBlock(image, where_x, where_y);
	else GraphicDevice::SetBlock(image, where_x, where_y);

	free(image);
}






void GraphicDevice::get_character_set_ptr(void)	{
	unsigned int char_seg;
	unsigned int char_off;

	asm push es;
	asm push bp;

	_AH=0x11;
	_AL=0x30;
	_BH=6;   	//	8x14 (2);	8x8 (3);	8x16 (6);
	geninterrupt(0x10);

	char_seg = _ES;
	char_off = _BP;
	character_set_ptr = (char far *)MK_FP(char_seg, char_off);

	asm pop bp;
	asm pop es;

}
