#ifndef __GRAPHICDEVICE
#define __GRAPHICDEVICE

#include "RGB.hpp"
#include <stdarg.h>

// WARNING! when using function Write(char *, ...)
#define MAX_CHAR_IN_WRITE   300 // WARNING: attempt to write more
																// than MAX_CHAR_IN_WRITE caracters
																// will be disastrous

#define BLOCK_SIZE(DX,DY) (4 + ((unsigned int)(DX) * (unsigned int)(DY)) * 3)
#define CHAR_WIDTH   8
#define CHAR_HEIGHT 16


// types of text colour gradient
enum enum_text_gradient {
	SOLID,
	GRADIENT_NS,
	GRADIENT_OE,
	GRADIENT_NO_SE,
	GRADIENT_SO_NE
};

// base class
class GraphicDevice {
	public:

		int XMax,YMax;

		int where_x, where_y;

		struct st_text_attr {
			int colour_type;   // types:  SOLID, GRADIENT_NS, GRADIENT_OE, GRADIENT_NO_SE, GRADIENT_SO_NE

			RGB colour;
			RGB colour_from;
			RGB colour_to;

			RGB paper_colour;
			RGB paper_colour_from;
			RGB paper_colour_to;
		} text;

	protected:
		static char far *character_set_ptr;
		static unsigned int ScanBufferBytesPerPixel;
		int print_on_screen;
	public:

	//----------------------------------------------------------------------------
		GraphicDevice(void);
		~GraphicDevice();

		// to be called after initialisation of derived clases
		// scan_buffer will be allocated with larger size if neccesary
		void ExtendScanBuffer(unsigned int x_size);

		/*------------------------------------------------------------------------
				pure virtual functions
		------------------------------------------------------------------------*/
		virtual void SetPixelNoClip(int x, int y,	RGB &rgb)=0;
		virtual RGB GetPixelNoClip(int x, int y)=0;
		virtual void SetHLineNoClip(unsigned char *source, int x, int y, int dx)=0;
		virtual void GetHLineNoClip(unsigned char *dest, int x, int y, int dx)=0;

		/*------------------------------------------------------------------------
			Clipped calls to _NoClip functions
		------------------------------------------------------------------------*/
		void SetPixel(int x, int y, RGB &rgb);
		RGB GetPixel(int x, int y);
		void SetHLine(unsigned char *source, int x, int y, int dx);
		void GetHLine(unsigned char *dest, int x, int y, int dx);
		//---------------------------------------------------------------
		void Line(int x1, int y1, int x2, int y2, RGB& rgb);
		void HLine(int x0, int y0, int dx, RGB& rgb);
		void VLine(int x, int y, int dy, RGB& rgb);
		void Bar(int x, int y, int dx, int dy, RGB rgb);
    void Rectangle(int x, int y, int dx, int dy, RGB rgb);
		void Rectangle(RGB rgb);  // 0,0, XMax,YMax
		void Cls(RGB rgb=RGB(0,0,0));

		// the buffer size required can be calculated using the macro IMAGE_SIZE(DX,DY)
		void GetBlock(unsigned char *buffer, int x, int y, int dx, int dy);
		void SetBlock(unsigned char *buffer, int x, int y);

		// The calling instance is the destiny
		//  e.g. screen.CopyBlock(0,0, i1,0,0, 20,20)
		// Can be used this way: CopyBlock(x_dest, y_dest, GraphicDevice &gd_org);
		//  and it will copy the whole source image.
		void CopyBlock( int x_dest, int y_dest,
			GraphicDevice &gd_org,  int x_org=0,  int y_org=0,
			int dx=-1, int dy=-1); // default (dx,dy) = maximum possible block

		void CopyBlockZoom( int x_dest, int y_dest,
										GraphicDevice &gd_org,  int x_org,  int y_org,
										int dx, int dy, int zoom=1);


		// use as printf
		void PrintF(char *format, va_list argptr);

		// print without format (no '\n', '\r' or %i, %f,...)
		void PrintString(char *string);


		// current coordinates (X = 0...XRes, Y = 0...YRes)
		void GotoXY(int where_x, int where_y);   // use as gotoxy
		int WhereX(void);
		int WhereY(void);

		// types of text colour gradient
		void SetTextColourType(enum enum_text_gradient);
    int GetTextColourType(void);



	// PROTECTED FUNCIONS
	protected:

		void get_character_set_ptr(void);

		void copy_block_zoom_neg(int x_dest, int y_dest,  GraphicDevice &gd_org,
			int x_org, int y_org,   int dx, int dy,   int zoom);



};

#endif