#include "ImgFile.hpp"
#include "Message.hpp"
#include "Pool.hpp"

#include "Screen.hpp"


#include <string.h>
#include <conio.h>

extern Screen screen;


extern MessageHandler message;
extern Pool pool;
/*----------------------------------------------------------------------------

----------------------------------------------------------------------------*/
ImgFile::ImgFile(void) {

	file_name=NULL;
	fp=NULL;
	temporary_image=0;

}



ImgFile::~ImgFile() {
	Close();
}

/*----------------------------------------------------------------------------
															New

	if file_name=="" => temporal file with random name
----------------------------------------------------------------------------*/
int ImgFile::New(char *file_nameVal, int dx, int dy) {

	XMax = dx;
	YMax = dy;

	if(file_name!=NULL) {
		delete file_name;
		file_name=NULL;
	}

	if(file_nameVal[0]==0) {
		// create temporary file_name
		file_name = new char [13];
		tmpnam(file_name);
		int len=strlen(file_name);
		file_name[len-3]='T'; file_name[len-2]='G'; file_name[len-1]='A';
		temporary_image=1;
	} else {

	// store string file_name
		file_name = new char [strlen(file_nameVal)+1];
		strcpy(file_name, file_nameVal);
	}

	// if there is a file open for this ImgFile instance it will be closed first
	if(fp!=NULL) {
		Close();
	}

	// creates a new file
	if ((fp = fopen(file_name, "w+b")) == NULL) {
		message.Message(EXIT | TO_SCREEN | TO_FILE, "ImgFile::Create  ' cannot open file <%s>", file_name);
		//return ERROR;
	}
	read_write = READ_WRITE;

	// prepares the TGA header
	unsigned char header []={
	0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,1,1,1, 0x18, 0x20 };

	// image size information for TGA header
	header[12] = dx % 256;
	header[13] = dx / 256;
	header[14] = dy % 256;
	header[15] = dy / 256;

	// writes header
	if (fwrite (header, 1, 18, fp) != 18) {
		message.Message(TO_SCREEN | EXIT, "ImgFile::Create  'Can't write header to file <%s>'", file_name);
	}
	pixel0 = 0x12L;


	// make sure there is enough room in pool.buffer
	ExtendScanBuffer(XMax);
	// clear pool.buffer

	//Cls(RGB(125,200,255));

	unsigned int i;
	for(i=0; i<XMax*3; i++) pool.buffer[i]=0;

	// fill the file with black scans
	for(int y=0; y<dy; y++) {
		if(fwrite(pool.buffer, 1, dx*3, fp) != dx*3) {
			message.Message(TO_SCREEN | EXIT, "ImgFile::Create  'Can't write to file <%s>'", file_name);
		}
	}

	inverted = 0;

	return OK;

}



/*----------------------------------------------------------------------------
															Open
----------------------------------------------------------------------------*/
int ImgFile::Open(char *file_nameVal, enum enum_read_write mode, int preview) {

	// store name
	file_name = new char [strlen(file_nameVal)+1];
	strcpy(file_name, file_nameVal);

	// if there is a file open for this ImgFile instance it will be closed first
	if(fp!=NULL) {
		Close();
	}

	// open file
	if ((fp = fopen(file_name, "rb")) == NULL) {
		message.Message(EXIT | TO_SCREEN | TO_FILE, "ImgFile::Open  ' cannot open file <%s>", file_name);
		//return ERROR;
	}

	char last_char = file_name[strlen(file_name)-1];
	if(last_char == 'A'|| last_char == 'a') read_TGA_header();
	else if(last_char == 'P' || last_char == 'p') read_BMP_header();
//	else if(last_char == 'F') read_TIF_header();
	else {
		XMax = 10;
		YMax = 10;
		pixel0 = 0;
	}


	read_write = mode;

	GraphicDevice::ExtendScanBuffer(XMax);  // extend scan_buffer

	return OK;
}



/*----------------------------------------------------------------------------
											BMP header
----------------------------------------------------------------------------*/
void ImgFile::read_BMP_header(void) {
	// reads image size BMP
	fseek(fp, 0x12L, SEEK_SET);
	XMax = fgetc(fp);
	XMax += fgetc(fp)<<8;

	bytes_per_scan_line = XMax*3;
	screen.PrintF("bytes_per_scan_line = %i  ", bytes_per_scan_line);
	getch();

	//if(!(bytes_per_scan_line&1)) bytes_per_scan_line++;//=bytes_per_scan_line;
	bytes_per_scan_line += 2;
	bytes_per_scan_line &= -1;

	fseek(fp, 0x16L, SEEK_SET);
	YMax = fgetc(fp);
	YMax += fgetc(fp)<<8;

	inverted = 1;

	pixel0 = 0x36;


}


/*----------------------------------------------------------------------------
											TGA header
----------------------------------------------------------------------------*/

void ImgFile::read_TGA_header(void) {

	fseek(fp, 12L, SEEK_SET);
	XMax = fgetc(fp);
	XMax += fgetc(fp)<<8;

	bytes_per_scan_line = XMax*3;

	YMax = fgetc(fp);
	YMax += fgetc(fp)<<8;

	// reads if image is inverted (first bytes in file correspond to last scan in image)
	fgetc(fp);
	if(fgetc(fp) == 0x00) inverted = 1;
	else inverted = 0;

	pixel0 = 0x12L;
}


/*----------------------------------------------------------------------------
															 Close
----------------------------------------------------------------------------*/
void ImgFile::Close(void) {

	fclose(fp);

	if(temporary_image) unlink(file_name);

	if(file_name!=NULL) {
		delete file_name;
		file_name=NULL;
	}


}

/*----------------------------------------------------------------------------
															 SaveAs
----------------------------------------------------------------------------*/
int ImgFile::SaveAs(char *file_name) {

	temporary_image=0;

	return 0;
}



/*----------------------------------------------------------------------------
																 ReDim
----------------------------------------------------------------------------*/
int ImgFile::ReDim(int dx, int dy) {

	return 0;

}

/*----------------------------------------------------------------------------
																SetPixelNoClip
----------------------------------------------------------------------------*/
void ImgFile::SetPixelNoClip(int x, int y,	RGB &rgb) {

	if(read_write != READ_WRITE) {
		message.Message(TO_FILE, "Imgage::SetPixel 'Attempt to write in Write-Protected Image'");
		return;
	}

	// clip
	if(x<0 || x>=XMax || y<0 || y>=YMax) return;

	if(inverted) fseek(fp, pixel0 + (x + (YMax-y-1) * (long)XMax)*3, SEEK_SET);
	else fseek(fp, pixel0 + (x + (long)y * XMax)*3, SEEK_SET);

	fputc(rgb.blue, fp);
	fputc(rgb.green, fp);
	fputc(rgb.red, fp);

}

/*----------------------------------------------------------------------------
															GetPixelNoClip
----------------------------------------------------------------------------*/
RGB ImgFile::GetPixelNoClip(int x, int y) {
  RGB rgb;

	// clip
	if(x<0 || x>=XMax || y<0 || y>=YMax) return RGB(0,0,0);

	if(inverted) fseek(fp, pixel0 + (x + (YMax-y-1) * (long)XMax)*3, SEEK_SET);
	else fseek(fp, pixel0 + (x + (long)y * XMax)*3, SEEK_SET);

	rgb.blue=fgetc(fp);
	rgb.green=fgetc(fp);
	rgb.red=fgetc(fp);

	return rgb;
}

/*----------------------------------------------------------------------------
															SetHLineNoClip
----------------------------------------------------------------------------*/
void ImgFile::SetHLineNoClip(unsigned char *source, int x, int y, int dx) {

	if(read_write != READ_WRITE) {
		message.Message(TO_FILE, "Imgage::SetPixel 'Attempt to write in Write-Protected Image'");
		return;
	}

	if(inverted) fseek(fp, pixel0 + (x + (YMax-y-1) * (long)XMax)*3, SEEK_SET);
	else fseek(fp, pixel0 + (x + (long)y * XMax)*3, SEEK_SET);

	fwrite(source, 3, dx, fp);

}


/*----------------------------------------------------------------------------
															GetHLineNoClip
----------------------------------------------------------------------------*/
void ImgFile::GetHLineNoClip(unsigned char *dest, int x, int y, int dx) {

	if(inverted) fseek(fp, (long)pixel0 + ((long)x + ((long)YMax-(long)y-1L) * bytes_per_scan_line), SEEK_SET);
	else fseek(fp, pixel0 + (x + (long)y * XMax)*3, SEEK_SET);

	fread(dest, 3, dx, fp);

}





