#include "ImgFile.hpp"
#include "ImgXMS.hpp"
#include "Message.hpp"
#include "XMS.hpp"
#include "Pool.hpp"
#include "Clip.hpp"
#include "Progress.hpp"

#include <string.h>
#include <conio.h>

#include "Screen.hpp"

extern Screen screen;

extern MessageHandler message;
extern XMS xms;
extern Pool pool;
/*----------------------------------------------------------------------------

----------------------------------------------------------------------------*/
ImgXMS::ImgXMS(void) {

	handle=0;
	temporary_image=1;
	file_name_allocated=0;

}



ImgXMS::~ImgXMS() {
	Close();
}

/*----------------------------------------------------------------------------
															New

	if file_name=="" => temporal file with random name
----------------------------------------------------------------------------*/
int ImgXMS::New(char *file_nameVal, int dx, int dy) {


	if(file_name_allocated) {
		delete file_name;
		file_name_allocated=0;
	}

	if(file_nameVal[0]==0) {
		// create temporary name
		file_name = new char [13];
		file_name_allocated=1;
		tmpnam(file_name);
		int len=strlen(file_name);
		file_name[len-3]='T'; file_name[len-2]='G'; file_name[len-1]='A';
		temporary_image=1;
	} else {

	// store string file_name
		file_name = new char [strlen(file_nameVal)+1];
		file_name_allocated=1;
		strcpy(file_name, file_nameVal);
	}

	message.Message(TO_FILE,"ImgXMS::Create   'creating \"%s\" (%i,%i)", file_name, dx,dy);

	// allocates XMS space
	if(handle!=0) xms.Free(handle);
	if ((handle = xms.Allocate((dx*(dy*3L))/1024+1 ) ) == 0) {
		message.Message(EXIT | TO_SCREEN | TO_FILE, "ImgXMS::Create  'cannot allocate XMS memory for <%s>'", file_name);
	}

	read_write = READ_WRITE;

	XMax = dx;
	YMax = dy;

	Cls();

	//InitGraphicDevice();  // extend scan_buffer
	GraphicDevice::ExtendScanBuffer(XMax);

	return OK;

}



/*----------------------------------------------------------------------------
															Open
----------------------------------------------------------------------------*/
int ImgXMS::Open(char *file_nameVal, enum enum_read_write mode, int preview) {

	ImgFile imgfile;

	if(imgfile.Open(file_nameVal, READ_ONLY) == ERROR) return ERROR;

	if(New(file_nameVal, imgfile.XMax, imgfile.YMax) == ERROR) return ERROR;
	read_write = mode;

	Progress* progress=NULL;
	if(preview) {
		sprintf(pool.buffer, "Loading <%s>", file_nameVal);
		progress = new Progress(0, YMax, 50, (char*)pool.buffer, RGB_RED, RGB_BLUE);
	}						//int xVal, int yVal, int dxVal, int dyVal);
//	this->CopyBlock(0,0, imgfile, 0,0, XMax, YMax);

	for(int y=0; y < YMax; y++) {
		imgfile.GetHLineNoClip(pool.buffer, 0, y, XMax);
		this->SetHLineNoClip(pool.buffer, 0, y, XMax);
		if(preview) {
			screen.SetHLine(pool.buffer, 0, y, XMax);
			progress->Increment();
		}
		if(kbhit()) if(getch()==0x1b) y=YMax;
	}

	if(progress) delete progress;
	return OK;
}

/*----------------------------------------------------------------------------
															 Close
----------------------------------------------------------------------------*/
void ImgXMS::Close(void) {

	message.Message(TO_FILE, "ImgXMS::Close  'closing <%s>'", file_name);

	if(!temporary_image) {
		// save image in file
		ImgFile i;
		i.New(file_name, XMax, YMax);
		i.CopyBlock(0,0, *this, 0,0, XMax, YMax);
	}

	if(handle) xms.Free(handle);

	if(file_name_allocated) {
		//message.Message(TO_FILE, "ImgXMS::Close  delete file_name <%s>'", file_name);
		delete file_name;
		file_name_allocated=0;
	}


}

/*----------------------------------------------------------------------------
															 SaveAs
----------------------------------------------------------------------------*/
int ImgXMS::SaveAs(char *file_name) {

	temporary_image=0;

	return 0;
}



/*----------------------------------------------------------------------------
																 ReDim
----------------------------------------------------------------------------*/
int ImgXMS::ReDim(int dx, int dy) {

	return 0;

}

/*----------------------------------------------------------------------------
																SetPixelNoClip
----------------------------------------------------------------------------*/
void ImgXMS::SetPixelNoClip(int x, int y,	RGB &rgb) {

	if(read_write != READ_WRITE) {
		message.Message(TO_FILE, "ImgXMS::SetPixel 'Attempt to write in Write-Protected Image'");
		return;
	}

	// clip
	//if(x<0 || x>=XMax || y<0 || y>=YMax) return;

	unsigned long add=(x + (unsigned long)XMax * y) *3L;

	pool.buffer[0]=rgb.blue;
	pool.buffer[1]=rgb.green;
	pool.buffer[2]=rgb.red;

	if(xms.Move(pool.buffer, handle, add, 3)!=1) {
		message.Message(TO_FILE, "ImgXMS::SetPixel Move ERROR'");
	}

}

/*----------------------------------------------------------------------------
															GetPixelNoClip
----------------------------------------------------------------------------*/
RGB ImgXMS::GetPixelNoClip(int x, int y) {

	// clip
	//if(x<0 || x>=XMax || y<0 || y>=YMax) return RGB(0,0,0);

	unsigned long add=(x + (unsigned long)XMax * y) *3L;

	// move an even number of bytes is faster

	if(xms.Move(handle, add, pool.buffer, 4)!=1) {
		message.Message(TO_FILE, "ImgXMS::GetPixel Move ERROR'");
	}

	return RGB(pool.buffer[2], pool.buffer[1], pool.buffer[0]);

}


/*----------------------------------------------------------------------------
															SetHLineNoClip
----------------------------------------------------------------------------*/
void ImgXMS::SetHLineNoClip(unsigned char *source, int x, int y, int dx) {

	if(read_write != READ_WRITE) {
		message.Message(TO_FILE, "Imgage::SetPixel 'Attempt to write in Write-Protected Image'");
		return;
	}

	//CLIP_HLINE(x,y,dx,XMax,YMax);

/*
	if(y<0 || y>=YMax) return;  // G, H
	if(x<0 || x+dx>XMax) {
		// clip needed
		if(x<0) {            // A, B, C
			dx=dx+x; x=0;      //    B, C
			if(dx<=0) return;  // A
		}                    //
		if(x+dx>XMax) {      // C, E, F
			dx=XMax-x;         // C, E
			if(x>XMax) return; //       F
		}
	}
*/
	unsigned long add=(x + (unsigned long)XMax * y) *3L;

	if(xms.Move(source, handle, add, dx*3)!=1) {
		message.Message(TO_FILE, "ImgXMS::SetHLine Move ERROR'");
	}

}


/*----------------------------------------------------------------------------
															GetHLineNoClip
----------------------------------------------------------------------------*/
void ImgXMS::GetHLineNoClip(unsigned char *dest, int x, int y, int dx) {

	unsigned long add=(x + (((unsigned long)XMax) * y)) *3L;

	/*
	 CLIP               G-----
								+------------------+
				 A----  |                  |
						B---+---               |
						 C--+------------------+--
								|   D--------      |
								|              E---+---
								|                  |  F----
								+------------------+
										 H------

	if(y<0 || y>=YMax) return;  // G, H
	if(x<0 || x+dx>XMax) {
		// clip needed
		if(x<0) {            // A, B, C
			dx=dx+x; x=0;      //    B, C
			if(dx<=0) return;  // A
		}                    //
		if(x+dx>XMax) {      // C, E, F
			dx=XMax-x;         // C, E
			if(x>XMax) return; //       F
		}
	}
	*/

	if(xms.Move(handle, add, dest, dx*3)!=1) {
		message.Message(TO_FILE, "ImgXMS::GetHLine Move ERROR'");
	}

}


