#include "Progress.hpp"
#include "Screen.hpp"
#include "Window.hpp"
#include "SaveBack.hpp"

extern Screen screen;

Progress::Progress(	int minVal, int maxVal, int stepVal,
							char *titleVal, RGB &rgb_inkVal, RGB &rgb_paperVal,
							int xVal, int yVal, int dxVal, int dyVal) {

	min = minVal;
	max = maxVal;
	step = stepVal;
	current = minVal;
	current_step = minVal;
  mid_x=0;

	rgb_ink  = new RGB;
	*rgb_ink = rgb_inkVal;
	rgb_paper = new RGB;
	*rgb_paper = rgb_paperVal;


	#define DEFAULT_BAR_WIDTH 20
	if(xVal==-1) {
		x  = WIN_FRAME_WIDTH;
		dx = screen.XMax - 2*WIN_FRAME_WIDTH;
		y  = screen.YMax - DEFAULT_BAR_WIDTH - 2*WIN_FRAME_WIDTH;
		dy = DEFAULT_BAR_WIDTH - (2*WIN_FRAME_WIDTH);
		if(titleVal[0] ==0) {    //no title
			y  -= CHAR_HEIGHT - WIN_FRAME_WIDTH;
			dy += CHAR_HEIGHT - WIN_FRAME_WIDTH;
		}
	}
	else {
		x  = xVal;
		y  = yVal;
		dx = dxVal;
		dy = dyVal;
	}

	if(titleVal[0] !=0) {
		win = new Window(titleVal, x, y, dx, dy,
						NO_CLOSE | NO_RESIZE | NO_HSCROLL_BAR | NO_VSCROLL_BAR);
	}
	else {
		win = new Window(titleVal, x, y, dx, dy,
						NO_TITLE | NO_RESIZE | NO_HSCROLL_BAR | NO_VSCROLL_BAR);
	}

	savebackground = new SaveBackground(win->frame_x, win->frame_y, win->frame_dx, win->frame_dy);

	win->img.New("", dx,dy);
	win->img.Cls(*rgb_paper);
	win->Print();
	win->Refresh();

}

Progress::~Progress() {
	delete savebackground;
	delete win;
	delete rgb_paper;
	delete rgb_ink;

}

void Progress::Increment(void) {
	current++;
	if(current - current_step >= step) {
		Set(current);
		current_step +=step;
	}
	//current += step;
	//Set(current);
}

void Progress::Set(int new_current) {

	if(new_current > max) new_current = max;
	if(new_current < min) new_current = min;

	int last_mid_x = mid_x;
	mid_x = ((long)dx * (long)(new_current - min)) / ( max - min );

//	win->img.Bar(0,0,mid_x,dy,*rgb_ink);
	win->img.Bar(last_mid_x,0,mid_x-last_mid_x,dy,*rgb_ink);

	// if need to fill background
	if(new_current < current) win->img.Bar(mid_x,0,dx-mid_x,dy,*rgb_paper);

	win->Refresh();

	current = new_current;
}
