#include "Screen.hpp"
#include "Message.hpp"
#include "Mouse.hpp"

#include <stdio.h>
#include <process.h>
#include <dos.h>
#include <conio.h>
#include <mem.h>
#include <alloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

extern MessageHandler message;
extern Mouse mouse;

int Screen::is_graphic_mode=0;  // set to 0 when in text mode
int Screen::old_video_mode=0;		// to restore on CloseGrpah()


/*----------------------------------------------------------------------------
												Constructor / Destructor
----------------------------------------------------------------------------*/
Screen::Screen(void) {
	is_graphic_mode = 0;
	GraphicDevice::get_character_set_ptr();
}

Screen::~Screen(void) { }

/*----------------------------------------------------------------------------
												Init/Close Graphics
----------------------------------------------------------------------------*/
int Screen::InitGraph(int mode) {
	int result;

	message.Message(TO_FILE, "Screen::InitGraph  'Initialising Video Mode 0x%X'", mode);

////	old_video_mode = VESA::GetSVGAVideoMode();

	result = VESA::InitVideoMode(mode);
	if(result==VESA_OK) {

		is_graphic_mode=1;

		result = BGraph::InitBGraph();  // set up Win A/B Memory page indexes
		if(result==VESA_OK) {

			GraphicDevice::XMax = VESA::XRes;
			GraphicDevice::YMax = VESA::YRes;

			GraphicDevice::ScanBufferBytesPerPixel = VESA::BytesPerPixel;
			GraphicDevice::ExtendScanBuffer(GraphicDevice::XMax);  // alloc static buffer

			return result;
		}
	}
	//else

	message.Message(TO_FILE, "Screen::InitGraph  'Failed to initialise Video Mode 0x%X.'", mode);
	return result;
}



void Screen::CloseGraph(void) {


	//SetSVGAVideoMode(old_video_mode);
	is_graphic_mode = 0;

	_AH = 0x00;
	_AL = 0x03;
	geninterrupt(0x10);     // init text mode 03h

	message.Message(TO_FILE, "Closing Graphic Mode");

	is_graphic_mode = 0;

}


/*----------------------------------------------------------------------------
				Smart Print functions (graph/text video mode)
----------------------------------------------------------------------------*/
void Screen::PrintF(char *format, ...) {
	va_list ap;
	va_start(ap, format);

	print_on_screen = 1;
	if(is_graphic_mode) GraphicDevice::PrintF(format, ap);
	else {
		if(XMax &&YMax) gotoxy((where_x*80)/XMax+1, (where_y*80)/YMax+1);
		vprintf(format, ...);
	}
	va_end(ap);

	print_on_screen = 0;

}

void Screen::PrintString(char *string) {

	print_on_screen = 1;
	if(is_graphic_mode) GraphicDevice::PrintString(string);
	else {
		if(XMax &&YMax) gotoxy((where_x*80)/XMax+1, (where_y*80)/YMax+1);
		printf(string);
	}
	print_on_screen = 0;

}

/*----------------------------------------------------------------------------
							Mouse temporal Hide   Functions
----------------------------------------------------------------------------*/
// TESTED
void Screen::SetPixel(int x, int y, RGB &rgb) {
	mouse.TemporalHide(x,y,1,1);
	GraphicDevice::SetPixel(x,y,rgb);
	mouse.TemporalShow();
}

// NOT TESTED
RGB Screen::GetPixel(int x, int y) {
	mouse.TemporalHide(x,y,1,1);
	RGB rgb=GraphicDevice::GetPixel(x,y);
	mouse.TemporalShow();
	return rgb;
}

// NOT TESTED
void Screen::SetHLine(unsigned char *source, int x, int y, int dx) {
	mouse.TemporalHide(x,y,dx,1);
	GraphicDevice::SetHLine(source, x, y, dx);
	mouse.TemporalShow();
}

// NOT TESTED
void Screen::GetHLine(unsigned char *dest, int x, int y, int dx) {
	mouse.TemporalHide(x,y,dx,1);
	GraphicDevice::GetHLine(dest, x, y, dx);
	mouse.TemporalShow();
}
//---------------------------------------------------------------
// TESTED

void Screen::Line(int x1, int y1, int x2, int y2, RGB& rgb) {
	mouse.TemporalHide(x1,y1, x2-x1+1, y2-y1+1);
	GraphicDevice::Line(x1, y1, x2,y2, rgb);
	mouse.TemporalShow();
}

// TESTED
void Screen::HLine(int x, int y, int dx, RGB& rgb) {
	mouse.TemporalHide(x,y,dx,1);
	GraphicDevice::HLine(x, y, dx, rgb);
	mouse.TemporalShow();
}

// TESTED
void Screen::VLine(int x, int y, int dy, RGB& rgb) {
	mouse.TemporalHide(x,y,1,dy);
	GraphicDevice::VLine(x, y, dy, rgb);
	mouse.TemporalShow();
}

// TESTED
void Screen::Bar(int x, int y, int dx, int dy, RGB rgb) {
	mouse.TemporalHide(x,y, dx,dy);
	GraphicDevice::Bar(x,y, dx,dy, rgb);
	mouse.TemporalShow();
}

// TESTED
void Screen::Rectangle(int x, int y, int dx, int dy, RGB rgb) {
	mouse.TemporalHide(x,y, dx,dy);
	GraphicDevice::Rectangle(x,y, dx,dy, rgb);
	mouse.TemporalShow();
}

// NOT TESTED
void Screen::Rectangle(RGB rgb) {
	mouse.TemporalHide(0,0, XMax,YMax);
	GraphicDevice::Rectangle(rgb);
	mouse.TemporalShow();
}

// NOT TESTED
void Screen::Cls(RGB rgb) {
	mouse.TemporalHide(0,0, XMax,YMax);
	GraphicDevice::Cls(rgb);
	mouse.TemporalShow();
}

// NOT TESTED
void Screen::GetBlock(unsigned char *buffer, int x, int y, int dx, int dy) {
	mouse.TemporalHide(x,y, dx,dy);
	GraphicDevice::GetBlock(buffer, x, y, dx, dy);
	mouse.TemporalShow();
}

// NOT TESTED
void Screen::SetBlock(unsigned char *buffer, int x, int y) {
	mouse.TemporalHide(x,y, buffer[0]+(buffer[1]<<8), buffer[2]+(buffer[3]<<8));
	GraphicDevice::SetBlock(buffer, x, y);
	mouse.TemporalShow();
}

// TESTED
void Screen::CopyBlock( int x_dest, int y_dest,
	GraphicDevice &gd_org,  int x_org,  int y_org,
	int dx, int dy)
{
	if(dx==-1) { // default meanings
		// x_org=0; y_org=0; dx=gd_org.XMax; dy=gd_org.YMax;
		mouse.TemporalHide(x_dest, y_dest, gd_org.XMax, gd_org.YMax);
		GraphicDevice::CopyBlock(
									x_dest, y_dest, gd_org, 0, 0, gd_org.XMax, gd_org.YMax);
		//GotoXY(0,200); PrintF("TemporalHide x_dest=%i y_dest=%i dx=%i dy=%i ",x_dest,y_dest, gd_org.XMax,gd_org.YMax);
	}
	else {
		mouse.TemporalHide(x_dest, y_dest, dx, dy);
		GraphicDevice::CopyBlock(x_dest, y_dest, gd_org, x_org, y_org, dx, dy);
		//GotoXY(0,200); PrintF("TemporalHide x_dest=%i y_dest=%i dx=%i dy=%i ",x_dest,y_dest,dx,dy);
	}
	mouse.TemporalShow();
//	getch();
}



// NOT TESTED
void Screen::CopyBlockZoom( int x_dest, int y_dest,
	GraphicDevice &gd_org,  int x_org,  int y_org,
	int dx, int dy, int zoom)
{
	mouse.TemporalHide(x_dest, y_dest, dx, dy);
	GraphicDevice::CopyBlockZoom(x_dest, y_dest, gd_org, x_org, y_org, dx, dy, zoom);
	//GotoXY(x_dest, y_dest);
	//PrintF("x_dest=%i y_dest=%i x_org=%i y_org=%i dx=%i dy=%i zoom=%i ",
	//				x_dest, y_dest, x_org, y_org, dx, dy, zoom);

	mouse.TemporalShow();
}


