#include "VESA.hpp"
#include "Message.hpp"

#include <stdio.h>
#include <process.h>
#include <dos.h>
#include <conio.h>
#include <mem.h>
#include <alloc.h>
#include <string.h>


extern MessageHandler message;

// Initialisation of static class VESA members. (Necesary since they are 'static')
char *VESA::VESAVersion=NULL;							// pointer to a string
char *VESA::VideoCardInfo=NULL;						// ptr to string. Same as in VGAInfoBlock.OEMStringPtr

unsigned int VESA::VideoMode=0;           //
unsigned int VESA::XRes=0;			          // horizontal resolution
unsigned int VESA::YRes=0;				        // vertical resolution
unsigned int VESA::BytesPerScanLine=0;    //
unsigned int VESA::MaxScanLine=0;    			// maximum scan line
unsigned int VESA::BytesPerPixel=0;  	  	//
unsigned long VESA::WinGranularity=0;     // window granularity
unsigned long VESA::WinAPointer=0;   			// same as WinASegment (window A start segment)
unsigned long VESA::WinBPointer=0;   			// same as WinBSegment (window B start segment)
void far (*VESA::WinFuncAddr)(void)=NULL; // pointer to windor function

int VESA::WindowABModel = USE_WIN_A_AND_B;		// defines the way to use the windows A,B
				// case USE_WIN_A_AND_B:
				//	 Win A is for write(only), Win B is for read(only).
				//	 (This allows to have two different 'win_pos' (MemoryWindow))
				//
				// case USE_ONLY_WIN_A:
				//	 Win A is for write and read.
				//	 (Can't use Win B, so MemoryWindow() will use WinA instead)

/*----------------------------------------------------------------------------

		InitVideoMode(int mode, int clear)

----------------------------------------------------------------------------*/
int VESA::InitVideoMode(int mode, int clear) {
	static char *VESAVersionString="X.X";
	int result;

// Information of Function 0 (VgaInfoBlock) is not needed to initialise Video Mode
	struct st_VgaInfoBlock VgaInfoBlock;
	result = GetVgaInfoBlock(&VgaInfoBlock);
	if(result == VESA_ERROR) return VESA_ERROR;

	// get VESA Version string
	VESAVersion = VESAVersionString;
	VESAVersion[0]=(VgaInfoBlock.VESAVersion>>8) + '0';
	VESAVersion[2]=(VgaInfoBlock.VESAVersion & 0xFF) + '0';

	// get Video Card info string
	VideoCardInfo = new char [strlen((char far*)VgaInfoBlock.OEMStringPtr+1)];
	strcpy(VideoCardInfo, (char far*)VgaInfoBlock.OEMStringPtr);


	// get ModeInfoBlock
	struct st_ModeInfoBlock ModeInfoBlock;
	result = GetModeInfoBlock(mode, &ModeInfoBlock);   // Function 01
	if(result == VESA_ERROR) {
		//message.Message(TO_FILE|TO_SCREEN,"Video Mode not supported.\n");
		return VESA_ERROR;
	}
	// properly inits Video Mode
	result = SetSVGAVideoMode(mode, clear);  // Function 02
	if(result == VESA_ERROR) return VESA_ERROR;


	// reads info for 'object VESA' from 'struct ModeInfoBlock'
	VideoMode = mode;
	XRes = ModeInfoBlock.XResolution;         // horizontal resolution
	YRes = ModeInfoBlock.YResolution;         // vertical resolution
	BytesPerScanLine = ModeInfoBlock.BytesPerScanLine;    // bytes per scan line
	BytesPerPixel = (ModeInfoBlock.BitsPerPixel+1)/8;  	  //
	WinGranularity = 1024L*ModeInfoBlock.WinGranularity;  // window granularity
	WinAPointer = (unsigned long)MK_FP(ModeInfoBlock.WinASegment, 0x0000);   // window A start segment
	WinBPointer = (unsigned long)MK_FP(ModeInfoBlock.WinBSegment, 0x0000);   // window B start segment
	WinFuncAddr = ModeInfoBlock.WinFuncAddr;  // pointer to windor function


	message.Message(TO_FILE, "VESA::InitVideoMode  Res(%u,%u) BPSL(%u) BPP(%u) WinGr(%lu) Card(%s) ",
		XRes,YRes, BytesPerScanLine, ModeInfoBlock.BitsPerPixel, WinGranularity, VideoCardInfo);


	// determines if WinB is readable    (case "Cirrus Logic GD-54xx VGA")
	if(WinBPointer==NULL) {
		WinBPointer = WinAPointer;
		WindowABModel = USE_ONLY_WIN_A;  // the MemoryWindow is the same for writing and reading, and corresponds to window A
	}

	//Get value of 'MaxScanLine' and 'BytesPerScan'
//not reliable	GetLogicalScanLineLength();

	return VESA_OK;

}


/*----------------------------------------------------------------------------

							Function 00 - Return Super VGA Information

----------------------------------------------------------------------------*/
int VESA::GetVgaInfoBlock(struct st_VgaInfoBlock *VgaInfoBlock) {

	_ES = FP_SEG(VgaInfoBlock);		// ES:DI = Pointer to buffer
	_DI = FP_OFF(VgaInfoBlock);
	_AH = 0x4F;										// Super VGA support
	_AL = 0x00;										// Return Super VGA information

	geninterrupt(0x10);

	return handle_status(_AX, 0x0000);						// Output: AX = Status

}

/*----------------------------------------------------------------------------

					Function 01h - Return Super VGA mode information

This function returns information about a specific Super VGA video mode that was
returned by Function 0.  The function fills a mode information block structure
at the address specified by the caller.  The mode information block size is
maximum 256 bytes.

----------------------------------------------------------------------------*/
int VESA::GetModeInfoBlock(unsigned int mode, struct st_ModeInfoBlock *ModeInfoBlock) {
	union REGS regs;
	struct SREGS sregs;

	regs.x.cx = mode;										// Super VGA video mode (mode number must be one of those returned by Function 0)
	sregs.es = FP_SEG(ModeInfoBlock);		// ES:DI = Pointer to buffer
	regs.x.di = FP_OFF(ModeInfoBlock);
	regs.h.ah = 0x4F;										// Super VGA support
	regs.h.al = 0x01;										// Return Super VGA mode information

	int86x(0x10, &regs, &regs, &sregs);

	return handle_status(regs.x.ax, 0x0100);			// Output: AX = Status

}



/*----------------------------------------------------------------------------

							Function 02h - Set Super VGA video mode

This function initializes a video mode.  The BX register contains the mode to
set.  The format of VESA mode numbers is described in chapter 2.  If the mode
cannot be set, the BIOS should leave the video environment unchanged and return
a failure error code.

----------------------------------------------------------------------------*/
int VESA::SetSVGAVideoMode(int mode, int clear) {  // Function 2

	mode |= (clear<<15);

	_BX = mode;				//D0-D14 = Video mode
										//D15 = Clear memory flag
												// 0 = Clear video memory
												// 1 = Don't clear video memory

	_AH = 0x4F;        // Super VGA support
	_AL = 0x02;        // Set Super VGA video mode

	geninterrupt(0x10);

	return handle_status(_AX, 0x0200);

}







unsigned int VESA::GetSVGAVideoMode(void) {
	unsigned int video_mode;

	_AH = 0x4F;        // Super VGA support
	_AL = 0x03;        // Return Super VGA Video Mode

	geninterrupt(0x10);

	video_mode = _BX;

	handle_status(_AX, 0x0300);

	return video_mode;

}





/*----------------------------------------------------------------------------
						 Function 05h - CPU Video Memory Window Control

This function sets or gets the position of the specified window in the video
memory.  The function allows direct access to the hardware paging registers.
To use this function properly, the software should use VESA BIOS Function 01h
(Return Super VGA mode information) to determine the size, location, and
granularity of the windows.

				Input:  AH = 4Fh        Super VGA support
								AL = 05h        Super VGA video memory window control
								BH = 00h        Select Super VGA video memory window
								BL = Window number
												0 = Window A
												1 = Window B
								DX = Window position in video memory
										 (in window granularity units)

				Output: AX = Status
								(See notes below)


				Input:  AH = 4Fh        Super VGA support
								AL = 05h        Super VGA video memory window control
								BH = 01h        Return Super VGA video memory window
								BL = Window number
												0 = Window A
												1 = Window B

				Output: AX = Status
								DX = Window position in video memory
										 (in window granularity units)
								(See notes below)


----------------------------------------------------------------------------*/

int VESA::SetMemoryWindowAViaInterrupt(int window_position) {  // Function 5
	_DX = window_position;// Window position in video memory (in window granularity units)
	_BL = 0x00;      // window_number:
											 //	0 = Window A
											 //	1 = Window B
	_AH = 0x4F;        // Super VGA support
	_AL = 0x05;        // Super VGA video memory window control
	_BH = 0x00;        // Select Super VGA video memory window
	geninterrupt(0x10);
	return handle_status(_AX, 0x0500);
}


int VESA::SetMemoryWindowBViaInterrupt(int window_position) {  // Function 5
	_DX = window_position;// Window position in video memory (in window granularity units)
	_BL = 0x01;      // window_number:
											 //	0 = Window A
											 //	1 = Window B
	_AH = 0x4F;        // Super VGA support
	_AL = 0x05;        // Super VGA video memory window control
	_BH = 0x00;        // Select Super VGA video memory window
	geninterrupt(0x10);
	return handle_status(_AX, 0x0500);
}

int VESA::SetMemoryWindowViaInterrupt(int window_position) {  // Function 5

	if( SetMemoryWindowAViaInterrupt(window_position) == VESA_ERROR) return VESA_ERROR;

	if( SetMemoryWindowBViaInterrupt(window_position) == VESA_ERROR) return VESA_ERROR;

	return VESA_OK;
}

/*----------------------------------------------------------------------------

													SetMemoryWindow (no interrupt)

Function 05h is also directly accessible through a far call from the
application.  The address of the BIOS function may be obtained by using VESA
BIOS Function 01h, Return Super VGA mode information.  A field in the
ModeInfoBlock contains the address of this function.  Note that this function
may be different among video modes in a particular BIOS implementation, so the
function pointer should be obtained after each set mode.

In the far call version, no status information is returned to the application.
Also, the AX and DX registers will be destroyed.  Therefore, if AX and/or DX
must be preserved, the application must do so priot to making the far call.

The application must load the input arguments in BH, BL, and DX (for set window)
but does not need to load either AH or AL in order to use the far call version
of this function.

----------------------------------------------------------------------------*/
// WINDOWS A and B
void VESA::SetMemoryWindow(int window_position) {  // Function 5
	_DX = window_position;// Window position in video memory (in window granularity units)
	_BL = 0x00;      // window_number:
											 //	0 = Window A
											 //	1 = Window B
	_BH = 0x00;        // Select Super VGA video memory window
	(*WinFuncAddr)();  // calls directly the function (no interrupt)
	// no status returned from the far call version

// AGAIN for window B
	_DX = window_position;// Window position in video memory (in window granularity units)
	_BL = 0x01;      // window_number:
											 //	0 = Window A
											 //	1 = Window B
	_BH = 0x00;        // Select Super VGA video memory window
	(*WinFuncAddr)();  // calls directly the function (no interrupt)
	// no status returned from the far call version

}


// WINDOW A
void VESA::SetMemoryWindowA(int window_position) {  // Function 5
	_DX = window_position;// Window position in video memory (in window granularity units)
	_BL = 0x00;      // window_number:
											 //	0 = Window A
											 //	1 = Window B
	_BH = 0x00;        // Select Super VGA video memory window
	(*WinFuncAddr)();  // calls directly the function (no interrupt)
// no status returned from the far call version
}

// WINDOW B
void VESA::SetMemoryWindowB(int window_position) {  // Function 5

	if(WindowABModel == USE_ONLY_WIN_A) { SetMemoryWindowA(window_position); return; }

	_DX = window_position;// Window position in video memory (in window granularity units)
	_BL = 0x01;      // window_number:
											 //	0 = Window A
											 //	1 = Window B
	_BH = 0x00;        // Select Super VGA video memory window
	(*WinFuncAddr)();  // calls directly the function (no interrupt)
// no status returned from the far call version
}


/*----------------------------------------------------------------------------

						 Function 06h - Set/Get Logical Scan Line Length

Get returns the current Bytes-Per-Scan-Line.
Set returns the Bytes-Per-Scan-Line actually set.

This function sets or gets the length of a logical scan line.  This function
allows an application to set up a logical video memory buffer that is wider
than the displayed area.  Function 07h then allows the application to set the
starting position that is to be displayed.

				Input:  AH = 4Fh        Super VGA support
								AL = 06h        Logical Scan Line Length
								BL = 00h        Select Scan Line Length
								CX = Desired width in pixels

				Output: AX = Status
								BX = Bytes Per Scan Line
								CX = Actual Pixels Per Scan Line
								DX = Maximum Number of Scan Lines


				Input:  AH = 4Fh        Super VGA support
								AL = 06h        Logical Scan Line Length
								BL = 01h        Return Scan Line Length

				Output: AX = Status
								BX = Bytes Per Scan Line
								CX = Actual Pixels Per Scan Line
								DX = Maximum Number of Scan Lines

Notes
~~~~~
The desired width in pixels may not be achieveable because of VGA hardware
considerations.  The next larger value will be selected thta will accommodate
the desired number of pixels, and the actual number of pixels will be returned
in CX.  BX returns a value that, when added to a pointer into video memory, will
point to the next scan line.  For example, in a mode 13h this would be 320, but
in mode 12h this would be 80.  DX returns the number of logical scan lines based
upon the new scan line length and the total memory installed and useable in this
display mode.  This function is also valid in text modes.  In text modes, the
application should find out the current character cell width through normal BIOS
functions, multiply that times the desired number of characters per line, and
pass the value in the CX register.
----------------------------------------------------------------------------*/
unsigned int VESA::SetLogicalScanLineLength(unsigned int pixels_per_scan_line_desired) {   // Function 6
	union REGS regs;
//	struct SREGS sregs;

	regs.x.cx = pixels_per_scan_line_desired;										// Super VGA video mode (mode number must be one of those returned by Function 0)
	regs.h.ah = 0x4F;										// Super VGA support
	regs.h.al = 0x06;                   // Logical Scan Line Length
	regs.h.bl = 0x00;                   // Select Scan Line Length

	int86(0x10, &regs, &regs);


	if(handle_status(regs.x.ax, 0x0600) == VESA_OK ) {
		BytesPerScanLine = regs.x.bx;	// BX  Bytes Per Scan Line
		MaxScanLine = regs.x.dx; 			// DX  Maximum Number of Scan Lines
		return regs.x.cx;  						// CX  Actual Pixels Per Scan Line
	}

	else
		return BytesPerScanLine;  // On Error returns preset BytesPerScanLine value

}


unsigned int VESA::GetLogicalScanLineLength(void) {   // Function 6
	union REGS regs;

	regs.h.ah = 0x4F;										// Super VGA support
	regs.h.al = 0x06;                   // Logical Scan Line Length
	regs.h.bl = 0x01;                   // Return Scan Line Length

	int86(0x10, &regs, &regs);

	if(handle_status(regs.x.ax, 0x0601) == VESA_OK ) {
		BytesPerScanLine = regs.x.bx;	// BX  Bytes Per Scan Line
		MaxScanLine = regs.x.dx; 			// DX  Maximum Number of Scan Lines
		return regs.x.cx;  						// CX  Actual Pixels Per Scan Line
	}
	else
		return BytesPerScanLine;  // On Error returns preset BytesPerScanLine value

}





/*----------------------------------------------------------------------------

						 Function 07h - Set/Get Display Start

This function selects the pixel to be displayed in the upper left corner of the
display from the logical page.  This function can be used to pan and scroll
around logical screens that are larger than the displayed screen.  This function
can also be used to rapidly switch between two different displayed screens for
double buffered animation effects.
----------------------------------------------------------------------------*/
int VESA::SetDisplayStart(unsigned int first_pixel_in_scan_line, unsigned int first_displayed_scan_line) {

	_CX = first_pixel_in_scan_line;	 // First Displayed Pixel in Scan Line
	_DX = first_displayed_scan_line; // First Displayed Scan Line
	_AH = 0x4F;         // Super VGA support
	_AL = 0x07;         // Display Start Control
	_BH = 0x00;         // Reserved and must be 0
	_BL = 0x00;         // Select Display Start

	geninterrupt(0x10);

	return handle_status(_AX, 0x0700);

}

int VESA::GetDisplayStart(unsigned int *first_pixel_in_scan_line,
unsigned int *first_displayed_scan_line)
{
	_AH = 0x4F;         // Super VGA support
	_AL = 0x07;         // Display Start Control
	_BL = 0x01;         // Return Display Start

	geninterrupt(0x10);

	*first_pixel_in_scan_line  = _CX;
	*first_displayed_scan_line = _DX;   // First Displayed Scan Line
	// BH = 00h Reserved and will be 0
	return handle_status(_AX, 0x0701);

}



/*----------------------------------------------------------------------------

							 Function 08h - Set/Get DAC Palette Control

This function queries and selects the operating mode of the DAC palette.  Some
DACs are configurable to provide 6-bits, 8-bits, or more of color definition per
red, green, and blue primary color.  The DAC palette width is assumed to be
reset to standard VGA 6-bits per primary during a standard or VESA Set Super VGA
Mode (AX = 4F02h) call.
----------------------------------------------------------------------------*/
int VESA::SetDACPalControl(unsigned char pal_width) {

	_AH = 0x4F;        // Super VGA support
	_AL = 0x08;        // Set/Get DAC Palette Control
	_BL = 0x00;        // Set DAC palette width
	_BH = pal_width;   // Desired number of bits of color per primary
										 // (Standard VGA = 6)
	geninterrupt(0x10);

	//Output:	BH = Current number of bits of color per primary (Standard VGA = 6)
	return handle_status(_AX, 0x0800);
}



unsigned char VESA::GetDACPalControl(void) {

	unsigned char pal_width;

	_AH = 0x4F;        // Super VGA support
	_AL = 0x08;        // Set/Get DAC Palette Control
	_BL = 0x01;        // Get DAC palette width

	geninterrupt(0x10);

	pal_width = _BH; // Current number of bits of color per primary (Standard VGA = 6)

	handle_status(_AX, 0x0801);
	return pal_width;

}




/*----------------------------------------------------------------------------
					 Returns VESA_ERROR or VESA_OK  depending on Status
----------------------------------------------------------------------------*/
int VESA::handle_status(int status, unsigned int function) {

	if(status == 0x004F) return VESA_OK;

		//	AL == 4Fh 	=>     Function is supported
		//	AL != 4Fh		=>     Function is not supported
		//	AH == 00h 	=>     Function call successful
		//	AH == 01h 	=>     Function call failed

	if( (status & 0x00FF) != 0x4F) message.Message(TO_FILE | BEEP, "VESA::handle_status   Function %4.4X is not supported\n", function);

	if( (status & 0xFF00) != 0x01) message.Message(TO_FILE | BEEP, "VESA::handle_status   Function(%4.4X) call failed", function);

	//sound(2000); delay(100);
	//sound(1000); delay(200);
	//nosound(); delay(160);

	return VESA_ERROR;

}



void VESA::InitTextMode(void) {

	_AH = 0x00;
	_AL = 0x03;
	geninterrupt(0x10);

}


