/*----------------------------------------------------------------------------
:                                                                            :
:										VESA Super VGA Standard VS911022.                        :
:											Programmed by Manuel Guillen.                          :
:						Based on Information available in PGCPE: VESA SVGA.              :
:                                                                            :
----------------------------------------------------------------------------*/
#ifndef __VESA
#define __VESA


#define VESA_ERROR   -1
#define VESA_OK       0

#define USE_WIN_A_AND_B   0
#define USE_ONLY_WIN_A    1

/*----------------------------------------------------------------------------
							'VgaInfoBlock'  struct definition
----------------------------------------------------------------------------*/
struct st_VgaInfoBlock {

	unsigned char VESASignature[4];    //db   'VESA'     ; 4 signature bytes
	unsigned int  VESAVersion;     		 //dw   ?          ; VESA version number
	unsigned char far*OEMStringPtr;    //dd   ?          ; Pointer to OEM string
	unsigned char Capabilities[4];     //db   4 dup(?)   ; capabilities of the video environment
	unsigned int far *VideoModePtr;    //dd   ?          ; pointer to supported Super VGA modes
	unsigned int  TotalMemory;         //dw   ?          ; Number of 64kb memory blocks on board
	unsigned char Reserved[236];       //db   236 dup(?) ; Remainder of VgaInfoBlock

};




/*----------------------------------------------------------------------------
							'ModeInfoBlock'  struct definition
----------------------------------------------------------------------------*/
struct st_ModeInfoBlock {

	// mandatory information

	unsigned int    ModeAttributes;      //dw  ?  ; mode attributes
	unsigned char   WinAAttributes;      //db  ?  ; window A attributes
	unsigned char   WinBAttributes;      //db  ?  ; window B attributes
	unsigned int    WinGranularity;      //dw  ?  ; window granularity
	unsigned int    WinSize;             //dw  ?  ; window size
	unsigned int    WinASegment;         //dw  ?  ; window A start segment
	unsigned int    WinBSegment;         //dw  ?  ; window B start segment
	void far(*WinFuncAddr)(void);        //dd  ?  ; pointer to windor function
	unsigned int    BytesPerScanLine;    //dw  ?  ; bytes per scan line

	// formerly optional information (now mandatory)

	unsigned int    XResolution;         //dw  ?  ; horizontal resolution
	unsigned int    YResolution;         //dw  ?  ; vertical resolution
	unsigned char   XCharCellSize;       //db  ?  ; character cell width
	unsigned char   YCharCellSize;       //db  ?  ; character cell height
	unsigned char   NumberOfPlanes;      //db  ?  ; number of memory planes
	unsigned char   BitsPerPixel;        //db  ?  ; bits per pixel
	unsigned char   NumberOfBanks;       //db  ?  ; number of banks
	unsigned char   MemoryModel;         //db  ?  ; memory model type
	unsigned char   BankSize;            //db  ?  ; bank size in kb
	unsigned char   NumberOfImagePages;  //db  ?  ; number of images
	unsigned char   Reserved0;           //db  1  ; reserved for page function

	//new Direct Color fields

	unsigned char   RedMaskSize;         //db  ?  ; size of direct color red mask in bits
	unsigned char   RedFieldPosition;    //db  ?  ; bit position of LSB of red mask
	unsigned char   GreenMaskSize;       //db  ?  ; size of direct color green mask in bits
	unsigned char   GreenFieldPosition;  //db  ?  ; bit position of LSB of green mask
	unsigned char   BlueMaskSize;        //db  ?  ; size of direct color blue mask in bits
	unsigned char   BlueFieldPosition;   //db  ?  ; bit position of LSB of blue mask
	unsigned char   RsvdMaskSize;        //db  ?  ; size of direct color reserved mask in bits
	unsigned char   DirectColorModeInfo; //db  ?  ; Direct Color mode attributes
	unsigned char   Reserved1[216];      //db  216 dup(?)  ; remainder of ModeInfoBlock

};



/*----------------------------------------------------------------------------
												Class     VESA
----------------------------------------------------------------------------*/

class VESA {

	public:

		static char *VESAVersion;
		static char *VideoCardInfo;							// same as in VGAInfoBlock.OEMStringPtr

		static unsigned int VideoMode;          //
		static unsigned int XRes;         			// video horizontal resolution
		static unsigned int YRes;         			// video vertical resolution
		static unsigned int BytesPerScanLine;   //
		static unsigned int MaxScanLine;				// maximum scan line
		static unsigned int BytesPerPixel;  	  //
		static unsigned long WinGranularity;    // window granularity in bytes
		static unsigned long WinAPointer;       // same as WinASegment (window A start segment)
		static unsigned long WinBPointer;       // same as WinBSegment (window B start segment)
		static void far (*WinFuncAddr)(void);   // pointer to MemoryWindow function
		// All members and functions are 'static' since there must be
		//   only one instance of this class.

		static int WindowABModel;		// defines the way to use the windows A,B
				// case USE_WIN_A_AND_B:
				//	 Win A is for write(only), Win B is for read(only).
				//	 (This allows to have two different 'win_pos' (MemoryWindow))
				//
				// case USE_ONLY_WIN_A:
				//	 Win A is for write and read.
				//	 (Can't use Win B, so MemoryWindow() will use WinA instead)

/*----------------------------------------------------------------------------
										CONTRUCTOR & DESTRUCTOR
----------------------------------------------------------------------------*/
	public:
//		VESA(void) {}
//		~VESA(void) {}

/*----------------------------------------------------------------------------
										Initialising Video Modes
----------------------------------------------------------------------------*/
	public:
		// Initialises a particular VESA Video Mode.
		// Returns VESA_OK or VESA_ERROR.
		int InitVideoMode(int mode, int clear=0);

		// Initialises Text Mode
		void InitTextMode(void);

/*----------------------------------------------------------------------------
						Extended VGA Bios.   FUNCTIONS  00h - 08h
----------------------------------------------------------------------------*/
	protected:
public:// for debug
	/*------------------------------------------------------------------------
				Function 00 - Get Super VGA Information
	------------------------------------------------------------------------*/
		// Fills VgaInfoBlock with useful Info.
		// (See 'struct st_VgaInfoBlock' definition below)
		// Returns VESA_OK or VESA_ERROR.
		int GetVgaInfoBlock(struct st_VgaInfoBlock *);


	/*------------------------------------------------------------------------
				Function 01 - Get Super VGA mode information
	------------------------------------------------------------------------*/
		// Fills ModeInfoBlock with useful Info.
		// (See 'struct st_ModeInfoBlock' definition below)
		// Returns VESA_OK or VESA_ERROR.
		int GetModeInfoBlock(unsigned int mode, struct st_ModeInfoBlock *);


	/*------------------------------------------------------------------------
				Function 02 - Set Super VGA mode
	------------------------------------------------------------------------*/
		// Performs a Low-Level initialisaton of VideoMode.
				// clear = 0   => Clear video memory
				// clear = 1   => Don't clear video memory
		// Returns VESA_OK or VESA_ERROR.
		int SetSVGAVideoMode(int mode, int clear=0);


	/*------------------------------------------------------------------------
				Function 03h - Return Super VGA Video Mode
	------------------------------------------------------------------------*/
		unsigned int GetSVGAVideoMode(void);


	/*------------------------------------------------------------------------
				Function 04h - Save/restore Super VGA video state
	------------------------------------------------------------------------*/
			// not impemented in this class


	/*------------------------------------------------------------------------
				Function 05: Super VGA video memory window control
	------------------------------------------------------------------------*/
		// To change the Memory Window use the faster SetMemoryWindow(int).
		// The memory window control via interrupt is slower.

		// No return value. Faster.
		void SetMemoryWindowA(int window_position); // window A
		void SetMemoryWindowB(int window_position); // window B
		void SetMemoryWindow(int window_position);  // window A and B

		// Returns VESA_OK or VESA_ERROR. Slow.
		int SetMemoryWindowAViaInterrupt(int window_position); // window A
		int SetMemoryWindowBViaInterrupt(int window_position); // window B
		int SetMemoryWindowViaInterrupt(int window_position);  // windows A and B


	/*------------------------------------------------------------------------
				Function 06: Set/Get Logical Scan Line Length (in pixels)
FUNCTIONS NOT RELIABLE: returns OK when it shouldn't.
											 so the BytesPerScanLine is set with nosense value!
	------------------------------------------------------------------------*/

		// Set returns the Pixels-Per-Scan-Line actually set.
		unsigned int SetLogicalScanLineLength(unsigned int pixels_per_scan_line_desired);

		// Get returns the current Pixels-Per-Scan-Line.
		unsigned int GetLogicalScanLineLength(void);


	/*------------------------------------------------------------------------
				Function 07: Set/Get Display Start
	------------------------------------------------------------------------*/
		// TO BE TESTED IN MODES 112,10F
		// FUNCTION NOT TESTED

		// Returns VESA_OK or VESA_ERROR.
		int SetDisplayStart(unsigned int first_pixel_in_scan_line,
			unsigned int first_displayed_scan_line);

		// Returns VESA_OK or VESA_ERROR.
		int GetDisplayStart(unsigned int *first_pixel_in_scan_line,
			unsigned int *first_displayed_scan_line);

	/*------------------------------------------------------------------------
				Function 08: Set/Get DAC Pal Control
	------------------------------------------------------------------------*/

		// Returns VESA_OK or VESA_ERROR.
		int SetDACPalControl(unsigned char pal_width);

		// returns the current number of bits of color per primary
		// (Standard VGA = 6)
		unsigned char GetDACPalControl(void);

	/*------------------------------------------------------------------------
				handle_status
	------------------------------------------------------------------------*/
	private:

		// Returns VESA_ERROR or VESA_OK  depending on status.
		// ('status' usually contains the value of register AX returned
		//   from Function calls)
		int handle_status(int status, unsigned int function);


};






#endif