#include "Window.hpp"
#include "Screen.hpp"
#include <string.h>

extern Screen screen;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Window::Window(char *title, int xVal, int yVal, int dxVal, int dyVal,
	WINDOW_ATTR attrVal)
{
	x = xVal;
	y = yVal;

	attr = attrVal;

	title_allocated = 0;
	SetTitle(title);

	Resize(dxVal, dyVal);


	//title_x = 0;

	shown.x = shown.y = 0;
	shown.zoom = 1;

	frame_colour = RGB(190,190,190);
	background_colour = RGB(90,90,90);
	title_colour = RGB(255,255,255);
	title_paper_colour = RGB(70,70,70);

	hyperWin = new Hyper(3);
	define_hyper();

}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Window::~Window() {
	if(title_allocated) delete title;
	delete hyperWin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::Print(void) {

	if(!(attr & NO_FRAME)) {

		if(!(attr & NO_TITLE)) {
			// frame with title
			screen.Rectangle(x-1,y-2-CHAR_HEIGHT, dx+2,dy+3+CHAR_HEIGHT, frame_colour);
			screen.Rectangle(x-2,y-3-CHAR_HEIGHT, dx+4,dy+5+CHAR_HEIGHT, title_paper_colour);
			screen.Rectangle(x-3,y-4-CHAR_HEIGHT, dx+6,dy+7+CHAR_HEIGHT, frame_colour);
			screen.HLine(x, y-1, dx, frame_colour);
		}
		else {
			// frame no title
			screen.Rectangle(x-1,y-1, dx+2,dy+2, frame_colour);
			screen.Rectangle(x-2,y-2, dx+4,dy+4, title_paper_colour);
			screen.Rectangle(x-3,y-3, dx+6,dy+6, frame_colour);
		}
	}

	//Refresh();  // ?????????????????????????????????????????????

	if(!(attr & NO_TITLE)) {

		// title text set up
		screen.text.colour = title_colour;
		screen.text.paper_colour = title_paper_colour;
		screen.text.colour_type = SOLID;

		// fill title background
		screen.Bar(x, y-CHAR_HEIGHT-1, dx, CHAR_HEIGHT, title_paper_colour);

//		screen.GotoXY(0,0);
//		screen.PrintF("title_lengh = %i  title_x = %i ", title_length, title_x);
		// print title text
		screen.GotoXY(x + title_x, y-CHAR_HEIGHT-1);


		char aux = title[title_length];
		title[title_length]=0;
		screen.PrintString(title);
		title[title_length] = aux;


		// 'close window' icon
#define   WIN_CLOSE_BUTTON_WIDTH   CHAR_WIDTH
		if(!(attr & NO_CLOSE)) {
			screen.VLine(x+dx-CHAR_WIDTH-1, y-CHAR_HEIGHT-1, CHAR_HEIGHT, frame_colour);
			screen.Bar(x+dx-(3*CHAR_WIDTH)/4, y-(3*CHAR_HEIGHT)/4-1, CHAR_WIDTH/2,
				CHAR_HEIGHT/2,title_colour);
		}
	}


}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::Refresh(void) {
	float mod_zoom = shown.zoom;
	if(shown.zoom == 0 ) mod_zoom = 1;
	if(shown.zoom < 0) mod_zoom = 1.0 / -mod_zoom;
	int aux_x;
	int aux_dx;
	screen.CopyBlockZoom(x,y, img, shown.x,shown.y, dx,dy, shown.zoom);

	if(!(attr & NO_HSCROLL_BAR)) {
		// Horizontal Scroll Bar
		aux_x = (shown.x * (long)dx) / img.XMax;
		aux_dx = (dx * (long)dx) / img.XMax / mod_zoom;
		if(aux_dx<=1) aux_dx=1;
		// clip scroll bar
		if(aux_x < 0) {
			aux_dx = aux_dx + aux_x; if(aux_dx<0) aux_dx=0;
			aux_x=0;
		}
		if(aux_x + aux_dx > dx) {
			aux_dx = dx-aux_x;
			if(aux_dx < 0) aux_dx=0;
		}
		// delete last scroll bar
		screen.HLine(x, y+dy+1, dx, RGB(0,0,200));
		// draws new scroll bar
		screen.HLine(
			x + aux_x,  // x + (shown.x * (long)dx) / img.XMax,
			y+dy+1,
			aux_dx,     // (dx * dx) / img.XMax / mod_zoom,
			RGB(255,255,0)
		);
	}

	if(!(attr & NO_VSCROLL_BAR)) {
		// vertical scroll bar
		aux_x = (shown.y * (long)dy) / img.YMax;
		aux_dx = (dy * (long)dy) / img.YMax / mod_zoom;
		if(aux_dx<=1) aux_dx=1;
		// clip scroll bar
		if(aux_x < 0) {
			aux_dx = aux_dx + aux_x; if(aux_dx<0) aux_dx=0;
			aux_x=0;
		}
		if(aux_x + aux_dx > dy) {
			aux_dx = dy-aux_x;
			if(aux_dx < 0) aux_dx=0;
		}
		screen.VLine(x+dx+1, y, dy, RGB(0,0,200));
		screen.VLine(
			x+dx+1,
			y + aux_x,    // (shown.y * (long)dy) / img.YMax,
			aux_dx,       // (dy * dy) / img.YMax / mod_zoom,
			RGB(255,255,0)
		);
	}

}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::SetTitle(char *titleVal) {

	if(title_allocated) delete title;
	title_allocated=0;

	title = new char [strlen(titleVal)+1];
	title_allocated=1;


	strcpy(title, titleVal);

}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::Resize(int dxVal, int dyVal) {

	float mod_zoom = shown.zoom;
	if(shown.zoom == 0 ) mod_zoom = 1;
	if(shown.zoom < 0) mod_zoom = 1.0 / -mod_zoom;

	if(dxVal == -1) {  // default value. It means to use the size of the image
		dx = img.XMax * mod_zoom;
		dy = img.YMax * mod_zoom;
	}
	else {
		dx = dxVal;
		dy = dyVal;
	}

	if(dx+x+3 > screen.XMax) dx = screen.XMax-3-x;
	if(dy+y+3 > screen.YMax) dy = screen.YMax-3-y;

	// calc title coordinate x in window's header
	title_x = ( (dx-((attr&NO_CLOSE)? 0:WIN_CLOSE_BUTTON_WIDTH))
					- ((signed)(strlen(title)+1)*CHAR_WIDTH) ) /2;

	if(title_x<0) {
		title_length = strlen(title) - (-2*title_x+CHAR_WIDTH)/CHAR_WIDTH;
		title_x = CHAR_WIDTH/2;
	} else title_length = strlen(title);

	frame_x = x - WIN_FRAME_WIDTH;
	frame_y = y -1 - WIN_FRAME_WIDTH - ((attr & NO_TITLE)? 0 : CHAR_HEIGHT);

	frame_dx = dx + 2* WIN_FRAME_WIDTH;
	frame_dy = dy +1 + 2* WIN_FRAME_WIDTH + ((attr & NO_TITLE)? 0 : CHAR_HEIGHT);


}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::Move(int xVal, int yVal) {
	x = xVal;
	y = yVal;

}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Window::define_hyper(void) {

	// shown area
	hyper_display= hyperWin->Define(x, y, dx, dy);

	// title
	hyper_title = hyperWin->Define(x, y-1-CHAR_HEIGHT, dx-CHAR_WIDTH-1, CHAR_HEIGHT);

	// close button
	hyper_close = hyperWin->Define(x + dx - CHAR_WIDTH, y-1-CHAR_HEIGHT, CHAR_WIDTH, CHAR_HEIGHT);

}


void Window::print_hyper(int x,int y) {

	screen.PrintF("hyper=2.2%i", hyperWin->GetHyper(x,y));

}
