#ifndef __WINDOW
#define __WINDOW


#include "XMS_USE.dir"

#if defined(_USE_XMS_)
	#include "ImgXMS.hpp"
#elif defined(_NO_USE_XMS_)
	#include "ImgFile.hpp"
#else
	#error You have to define _USE_XMS_ or _NO_USE_XMS in Options/Compiler/Code Generation
#endif

#include "Hyper.hpp"

#define WIN_FRAME_WIDTH  3

// window attr:
enum WINDOW_ATTR {
	ALL_FEATURES      = 0,
	NO_FRAME          = 1,
	NO_TITLE          = 2,
	NO_RESIZE         = 4,
	NO_HSCROLL_BAR    = 8,
	NO_VSCROLL_BAR    = 16,
	NO_CLOSE          = 32
};


class Window {

	public:


		// visible image size and position
		int x,y;   // image position (of shown area) on screen
		int dx,dy; //  and dimensions (of shown area)

		// window size and position
		int frame_x,frame_y;   // window position (of frame) on screen
		int frame_dx,frame_dy; //  and dimensions (of frame)

		#if defined(_USE_XMS_)
			ImgXMS img;
		#elif defined(_NO_USE_XMS_)
			ImgFile img;
		#endif

		struct st_shown {
			int x,y;      // image coordinates of first pixel of shown area
			int zoom;     // image zoom
		} shown;

		char *title;   // title of window
		char title_allocated;  // set when memory for title has been allocated
		int title_x;   // x coordinate of title text (allows the text to be centered)
		int title_length;  // number of characters that fits into window's header
		Hyper *hyperWin;
		HYPER hyper_title, hyper_close, hyper_display;

		unsigned int attr; // see define list at the beguining of this file.

		// colours
		RGB frame_colour;
		RGB background_colour;
		RGB title_colour;
		RGB title_paper_colour;



	public:

		Window(char *title, int xVal=0, int yVal=0, int dxVal=100, int dyVal=100,
						WINDOW_ATTR attr=ALL_FEATURES);
		~Window();

		void Refresh(void);
		void Print(void);
		void SetTitle(char *titleVal);
		void Resize(int dxVal=-1, int dyVal=-1);
		void Move(int xVal, int YVal);

		void print_hyper(int x, int y);

	private:
		void define_hyper(void);

};



#endif