// VERY IMPORTANT !
// The dest-address in Move() must fit in a segment, without
//  offset change from 0xFFFF to 0x0000
//  eg: &buf[3]=6F39:FFFF   &buf[4]=6F39:0000   Not Valid!
// This can be assured by allocating memory with 'malloc'
//  instead of by 'char buf[100]'



#include "XMS.hpp"
#include "Message.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>


extern MessageHandler message;

//------------------------------------------------------------------------
//   CONSTRUCTOR  &  DESTRUCTOR
//------------------------------------------------------------------------
XMS::XMS(void) {

	for(int index=0; index < MAX_N_HANDLES_IN_LIST; index++) {
		handle_list[index].used = 0;
	}

	if(is_driver_installed()) {
		//message.Message(TO_SCREEN | TO_FILE, "XMS::XMS    'XMS driver found'");
		get_function_address();
	}
	else {
		message.Message(EXIT | TO_FILE | TO_SCREEN | BEEP,"XMS::XMS   'XMS driver not installed!'");
	}

	//message.Message(TO_SCREEN|TO_FILE,
		//"XMS Version(%X) Free Memory(%uKb) Largest Free Block(%uKb) FreeHandles(%i)",
		//GetVersionNumber(), QueryFreeMemory(),
		//LargestFreeBlock(), FreeHandles());

}



XMS::~XMS() {
	FreeAll();    // cool!
}


//------------------------------------------------------------------------
//   SET UP AND INSTALLATION
//------------------------------------------------------------------------

// returns 1 if the driver is installed, 0 otherwise
int XMS::is_driver_installed(void) {
	_AH = 0x43;
	_AL = 0x00;
	geninterrupt(0x2F);
	return (_AL==0x80);
}

// Obtains the address of the driver's control function
void XMS::get_function_address(void) {
	unsigned int seg,off;

	_AH = 0x43;
	_AL = 0x10;
	geninterrupt(0x2F);

	seg = _ES;
	off = _BX;
	xms_function = (void (far *)())MK_FP(seg, off);

}


//------------------------------------------------------------------------
//   INFORMATION AND QUERIES
//------------------------------------------------------------------------

// For debug purposes. It shows the list of allocated handles
void XMS::PrintHandleList(void) {
	int index;

	for(index=0; index < MAX_N_HANDLES_IN_LIST; index++) {
		if(handle_list[index].used)
			printf("<%i>; handle <%X>;  Length <%u Kb>\n",
				index, handle_list[index].handle, Length(handle_list[index].handle));
	}

}


// returns 16-bit BCD number representing the XMS Version number
unsigned int XMS::GetVersionNumber(void) {

	_AH=0x00;
	(*xms_function)();
	return _AX;

}

// returns total free memory in Kb
unsigned int XMS::QueryFreeMemory(void) {

	_AH=0x08;
	(*xms_function)();
	return _DX;

}

// returns size of largest free block in Kb
unsigned int XMS::LargestFreeBlock(void) {

	_AH=0x08;
	(*xms_function)();
	return _AX;

}

// returns the number of free handles in the system
unsigned int XMS::FreeHandles(void) {

	int index;

	for(index=0; index < MAX_N_HANDLES_IN_LIST; index++) {
		if(handle_list[index].used)
		return(free_handles_subfunction(handle_list[index].handle));
	}

	unsigned int free_handles;
	XMS_HANDLE aux_handle;

	aux_handle = Allocate(0);

	free_handles = free_handles_subfunction(aux_handle) + 1;

	Free(aux_handle);

	return free_handles;

}

// returns the number of free handles in the system
unsigned int XMS::free_handles_subfunction(XMS_HANDLE valid_handle) {

	_DX=valid_handle;
	_AH=0x0E;
	(*xms_function)();
	return _BL;

}

// returns size in Kb allocated for that Handle.
unsigned int XMS::Length(XMS_HANDLE handle) {

	_DX=handle;
	_AH=0x0E;
	(*xms_function)();
	if(_AX!=0x0001) {
		message.Message(TO_FILE | TO_SCREEN,"XMS::Length  'Info of handle(%u) not found'", handle);
	}
	return _DX;

}

//------------------------------------------------------------------------
//   ALLOCATE, REALLOCATE AND FREE
//------------------------------------------------------------------------

// returns the handle value if the block is allocated, 0 otherwise
XMS_HANDLE XMS::Allocate(unsigned int size) { // size in Kb
	int index;
	XMS_HANDLE handle;

	for(index=0; index < MAX_N_HANDLES_IN_LIST; index++) {
		if(handle_list[index].used==0) {
			handle = allocate_subfunction(size);
			handle_list[index].handle = handle;

			if(handle) {
				handle_list[index].used=1;
				message.Message(TO_FILE,"XMS::Allocate  'allocated(%uKB)'", size);
			}
			return handle;
		}
	}


	message.Message(TO_FILE,"XMS::Allocate   'No more space in the list of handles'");
	return 0;
}

// returns the handle value if the block is allocated, 0 otherwise
XMS_HANDLE XMS::allocate_subfunction(unsigned int size) {
	XMS_HANDLE handle;

	_DX=size;
	_AH=0x09;
	(*xms_function)();
	if(_AX!=0x0001) {
		message.Message(TO_FILE, "XMS::allocate_subfunction   'CAN'T ALLOCATE'");
		return 0;
	}
	handle = _DX;
	return handle;

}

// returns 1 if the block is reallocated, 0 otherwise
unsigned int XMS::Reallocate(XMS_HANDLE handle, unsigned int new_size) {	// size in Kb
//# la function realocate de XMS no funciona correctamente.

//	asm push dx;
//	asm push bx;

	_DX=handle;
	_BX=new_size;
	_AH=0x0F;
	(*xms_function)();

//	if(_AX!=0x0001) {
//		printf("CAN'T REALLOCATE\n");
//	}

//	asm pop bx;
//	asm pop dx;

	return _AX;


}

// returns 1 if the block is freed, 0 otherwise
unsigned int XMS::Free(XMS_HANDLE handle) {
	int index;
	unsigned int result;

	for(index=0; index < MAX_N_HANDLES_IN_LIST; index++) {
		if(handle_list[index].handle == handle) {
			result = free_subfunction(handle);
			if(result) handle_list[index].used = 0;
			return result;
		}
	}

	message.Message(TO_FILE, "XMS::Free   'Can't find handle to be Freed'");
	return 0;
}




// returns 1 if the block is freed, 0 otherwise
unsigned int XMS::free_subfunction(XMS_HANDLE handle) {

	_DX=handle;
	_AH=0x0A;
	(*xms_function)();
	return _AX;

}

// Deallocates all memory blocks
// when exiting the program Destructor ~XMS will call this function anyway.
void XMS::FreeAll(void) {

	int index;

	for(index=0; index < MAX_N_HANDLES_IN_LIST; index++) {
		if(handle_list[index].used) Free(handle_list[index].handle);
	}

}

//------------------------------------------------------------------------
//   MOVE MEMORY    Length must be even !! !!
//------------------------------------------------------------------------

// Move Extended memory To Conventional memory  (length in bytes)
int XMS::Move(XMS_HANDLE source_handle, unsigned long source_offset,
void far *dest_address, unsigned long length) {

	move_struct.length = length;

	move_struct.source_handle = source_handle;
	move_struct.source_offset = source_offset;

	move_struct.dest_handle = 0x0000;
	move_struct.dest_offset = (unsigned long)dest_address;

	// Case ODD length
	// XMS Driver does not move odd length blocks. So its implemented in two parts
	if(length&1) {
		unsigned char buffer[2];
		int result;

//		printf("                          ODD MOVE BUFFER ADDRESS = %p\n", buffer);
		// Move length-1 bytes from EM to dest

			move_struct.length--;
			result = move_subfunction();
			if(result != 0x0001) return result;

		// Move 2 bytes from EM to buffer

			move_struct.source_offset += length-1;
			move_struct.dest_offset = (unsigned long) buffer;
			move_struct.length=2;
			result = move_subfunction();

		// Move 1 byte from buffer to array
			*((unsigned char*)(dest_address)+length-1) = buffer[0];

			return result;

	}



	return move_subfunction();
}

// Move Conventional memory To Extended memory (length in bytes)
int XMS::Move(void far *source_address, XMS_HANDLE dest_handle,
unsigned long dest_offset, unsigned long length) {

	move_struct.length = length;

	move_struct.source_handle = 0x0000;
	move_struct.source_offset = (unsigned long)source_address;

	move_struct.dest_handle = dest_handle;
	move_struct.dest_offset = dest_offset;

	// Case ODD length
	// XMS Driver does not move odd length blocks. So its implemented in two parts
	if(length&1) {

		if(length==1) {
			return 0xFF; // Can't move a single byte. Minimum 2 bytes
		}

		unsigned char buffer[2];
		int result;

//		printf("--------------------------ODD MOVE BUFFER ADDRESS = %p\n", buffer);
		// Move length-1 bytes from source to EM

			move_struct.length--;
			result = move_subfunction();
			if(result != 0x0001) return result;

		// Move last 2 bytes from source to EM

			move_struct.dest_offset += length-2;
			move_struct.source_offset = (unsigned long)source_address+length-2;
			move_struct.length=2;
			result = move_subfunction();

			return result;

	}


	return move_subfunction();
}

// Move Extended memory To Extended memory    (length in bytes)
int XMS::Move(XMS_HANDLE source_handle,unsigned long source_offset,
XMS_HANDLE dest_handle, unsigned long dest_offset, unsigned long length) {

	move_struct.length = length;

	move_struct.source_handle = source_handle;
	move_struct.source_offset = source_offset;

	move_struct.dest_handle = dest_handle;
	move_struct.dest_offset = dest_offset;

	return move_subfunction();
}

// Move Conventional memory To Conventional memory (length in bytes)
int XMS::Move(void far *source_address, void far *dest_address, unsigned long length) {

	move_struct.length = length;

	move_struct.source_handle = 0x0000;
	move_struct.source_offset = (unsigned long)source_address;

	move_struct.dest_handle = 0x0000;
	move_struct.dest_offset = (unsigned long)dest_address;

	return move_subfunction();
}

#include <conio.h>

int XMS::move_subfunction(void) {
	int ds,si;

/*
printf("length = %lu\nsource_handle = %X\nsource_offset = %lu\ndest_handle = %X\ndest_offset = %p\n",
	move_struct.length, move_struct.source_handle,
	move_struct.source_offset, move_struct.dest_handle,
	move_struct.dest_offset);
getch();
*/
	ds= _DS;
	si= _SI;

	_DS = FP_SEG(&move_struct);
	_SI = FP_OFF(&move_struct);
	_AH=0x0B;
	(*xms_function)();
//	if(_AX==0x0001) return 0x0001;
	int ax=_AX;
	_DS=ds;
	_SI=si;

	int bl =_BL;

//	printf("((ax=%X))\n", ax);
	if(ax!=0x0001) {
		message.Message(TO_FILE,"XMS::movesubfunction   'ERROR in Move: ");

		switch(bl) {
			case 0x80: message.Message(TO_FILE,"  Function not implemented"); break;
			case 0x81: message.Message(TO_FILE,"  VDISK device detected"); break;
			case 0x82: message.Message(TO_FILE,"  A20 error"); break;

			case 0xA3: message.Message(TO_FILE,"  Source Handle is invalid"); break;
			case 0xA4: message.Message(TO_FILE,"  Source Offset is invalid"); break;
			case 0xA5: message.Message(TO_FILE,"  Dest Handle is invalid"); break;
			case 0xA6: message.Message(TO_FILE,"  Dest Offset is invalid"); break;
			case 0xA7: message.Message(TO_FILE,"  Length(%lu) is invalid", move_struct.length); break;

			case 0xA8: message.Message(TO_FILE,"  Invalid overlap"); break;
			case 0xA9: message.Message(TO_FILE,"  Parity error"); break;

			default: message.Message(TO_FILE,"  Unknown return value from XMS Move  BL=%i", bl);
		}
		return bl;
	}

	return 0x0001;


}
