// Manuel Guillen 19/10/1996


/* ================ VERY IMPORTANT ! ========================================

1.-		IN FUNCTION Move()

				The dest_address array in Move() must fit in a segment, without
					offset change from 0xFFFF to 0x0000
					eg: &buf[3]=6F39:FFFF   &buf[4]=6F39:0000   Not Valid!

				This can be assured by allocating memory with 'malloc'
					instead of by 'char buf[100]'


2.-		IN FUNCTION Move()

				Move Conventional memory To Extended memory
				Can't move a single byte. Minimum 2 bytes (or 0 bytes)

========================================================================== */

/*----------------------------------------------------------------------------

		More than one instance of XMS can be created.

		Return value of function Move:

		SUCCESS

				0x01   Success. Otherwise error

		ERROR:

				0x80   Function not implemented
				0x81   VDISK device detected
				0x82   A20 error

				0xA3   Source Handle is invalid
				0xA4   Source Offset is invalid
				0xA5   Dest Handle is invalid
				0xA6   Dest Offset is invalid
				0xA7   Length is invalid

				0xA8   Invalid overlap
				0xA9   Parity error

				0xFF   Can't move a single byte  // from Conventional to Extended Mem.
																				 // Minimum 2 bytes (or 0 bytes)

----------------------------------------------------------------------------*/




#ifndef __XMS
#define __XMS

typedef unsigned int XMS_HANDLE;

#define MAX_N_HANDLES_IN_LIST 256

class XMS {

	private:
		void far (*xms_function)(void);//Pointer to the driver's control function

		struct st_allocated_handles {
			XMS_HANDLE handle;
			char used;
		} handle_list[MAX_N_HANDLES_IN_LIST];

		struct st_move {
			unsigned long length;
			unsigned int source_handle;
			unsigned long source_offset;
			unsigned int dest_handle;
			unsigned long dest_offset;
		} move_struct;


	private:
		// returns 1 if the block is allocated, 0 otherwise
		XMS_HANDLE allocate_subfunction(unsigned int size); // size in Kb

		// returns 1 if the block is freed, 0 otherwise
		unsigned int free_subfunction(XMS_HANDLE handle);

		// returns the number of free handles in the system
		unsigned int free_handles_subfunction(XMS_HANDLE valid_handle);

		int move_subfunction(void);


//	private:

public:		//------------------------------------------------------------------------
		// 	SET UP AND INSTALLATION	 (private)
		//------------------------------------------------------------------------

		// returns 1 if the driver is installed, 0 otherwise
		int is_driver_installed(void);

		// Obtains the address of the driver's control function
		void get_function_address(void);

		//------------------------------------------------------------------------
		//   CONSTRUCTORS
		//------------------------------------------------------------------------

	public:
		XMS(void);
		~XMS();

		//------------------------------------------------------------------------
		//   INFORMATION AND QUERIES
		//------------------------------------------------------------------------

		// debug only
    void PrintHandleList(void);


		// returns 16-bit BCD number representing the XMS Version number
		unsigned int GetVersionNumber(void);

		// returns total free memory in Kb
		unsigned int QueryFreeMemory(void);

		// returns size of largest free block in Kb
		unsigned int LargestFreeBlock(void);

		// returns the number of free handles in the system
		unsigned int FreeHandles(void);

		// returns size in Kb allocated for that Handle.
		unsigned int Length(XMS_HANDLE handle);

		//------------------------------------------------------------------------
		//   ALLOCATE, REALLOCATE AND FREE
		//------------------------------------------------------------------------

		// returns the handle value if the block is allocated, 0 otherwise
		XMS_HANDLE Allocate(unsigned int size); // size in Kb

		// returns 1 if the block is reallocated, 0 otherwise
		unsigned int Reallocate(XMS_HANDLE handle, unsigned int new_size); // size in Kb

		// returns 1 if the block is freed, 0 otherwise
		unsigned int Free(XMS_HANDLE handle);

		// Deallocates all memory blocks
		// when exiting the program Destructor ~XMS will call this function anyway.
		void FreeAll(void);


		//------------------------------------------------------------------------
		//   MOVE MEMORY
		//------------------------------------------------------------------------

		// Move Extended memory To Conventional memory  (length in bytes)
		int Move(XMS_HANDLE source_handle, unsigned long source_ofset,
		void far *dest_address, unsigned long length);

		// Move Conventional memory To Extended memory (length in bytes)
		int Move(void far *source_address, XMS_HANDLE dest_handle,
		unsigned long dest_offset, unsigned long length);

		// Move Extended memory To Extended memory    (length in bytes)
		int Move(XMS_HANDLE source_handle,unsigned long source_offset,
		XMS_HANDLE dest_handle, unsigned long dest_offset, unsigned long length);

		// Move Conventional memory To Conventional memory (length in bytes)
		int Move(void far *mem_source, void far *mem_dest, unsigned long length);


};




#endif